# Copyright (c) 2015 Alex Meade.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import time
from unittest import mock

import ddt
from oslo_config import cfg

from manila import exception
from manila.share import configuration
from manila.share import driver
from manila.share.drivers.netapp.dataontap.client import api as netapp_api
from manila.share.drivers.netapp.dataontap.client import client_cmode
from manila.share.drivers.netapp.dataontap.cluster_mode import data_motion
from manila.share.drivers.netapp import options as na_opts
from manila.share.drivers.netapp import utils as na_utils
from manila.share import utils as share_utils
from manila import test
from manila.tests.share.drivers.netapp.dataontap import fakes as fake
from manila.tests.share.drivers.netapp import fakes as na_fakes


CONF = cfg.CONF


@ddt.ddt
class NetAppCDOTDataMotionTestCase(test.TestCase):

    def setUp(self):
        super(NetAppCDOTDataMotionTestCase, self).setUp()
        self.backend = 'backend1'
        self.mock_cmode_client = self.mock_object(client_cmode,
                                                  "NetAppCmodeClient",
                                                  mock.Mock())
        self.config = configuration.Configuration(driver.share_opts,
                                                  config_group=self.backend)
        self.config.append_config_values(na_opts.netapp_cluster_opts)
        self.config.append_config_values(na_opts.netapp_connection_opts)
        self.config.append_config_values(na_opts.netapp_basicauth_opts)
        self.config.append_config_values(na_opts.netapp_transport_opts)
        self.config.append_config_values(na_opts.netapp_support_opts)
        self.config.append_config_values(na_opts.netapp_provisioning_opts)
        self.config.append_config_values(na_opts.netapp_data_motion_opts)
        CONF.set_override("share_backend_name", self.backend,
                          group=self.backend)
        CONF.set_override("netapp_transport_type", "https",
                          group=self.backend)
        CONF.set_override("netapp_login", "fake_user",
                          group=self.backend)
        CONF.set_override("netapp_password", "fake_password",
                          group=self.backend)
        CONF.set_override("netapp_server_hostname", "fake.hostname",
                          group=self.backend)
        CONF.set_override("netapp_server_port", 8866,
                          group=self.backend)
        CONF.set_override("netapp_ssl_cert_path", "/etc/ssl/certs",
                          group=self.backend)

    def test_get_client_for_backend(self):
        self.mock_object(data_motion, "get_backend_configuration",
                         mock.Mock(return_value=self.config))

        data_motion.get_client_for_backend(self.backend)

        self.mock_cmode_client.assert_called_once_with(
            hostname='fake.hostname', password='fake_password',
            username='fake_user', transport_type='https', port=8866,
            ssl_cert_path='/etc/ssl/certs', trace=mock.ANY, vserver=None)

    def test_get_client_for_backend_with_vserver(self):
        self.mock_object(data_motion, "get_backend_configuration",
                         mock.Mock(return_value=self.config))

        CONF.set_override("netapp_vserver", 'fake_vserver',
                          group=self.backend)

        data_motion.get_client_for_backend(self.backend)

        self.mock_cmode_client.assert_called_once_with(
            hostname='fake.hostname', password='fake_password',
            username='fake_user', transport_type='https', port=8866,
            ssl_cert_path='/etc/ssl/certs', trace=mock.ANY,
            vserver='fake_vserver')

    def test_get_client_for_host(self):
        mock_extract_host = self.mock_object(
            share_utils, 'extract_host',
            mock.Mock(return_value=fake.BACKEND_NAME))
        mock_get_client = self.mock_object(
            data_motion, 'get_client_for_backend',
            mock.Mock(return_value=self.mock_cmode_client))

        returned_client = data_motion.get_client_for_host(
            fake.HOST_NAME)

        mock_extract_host.assert_called_once_with(
            fake.HOST_NAME, level='backend_name')
        mock_get_client.assert_called_once_with(fake.BACKEND_NAME)
        self.assertEqual(returned_client, self.mock_cmode_client)

    def test_get_config_for_backend(self):
        self.mock_object(data_motion, "CONF")
        CONF.set_override("netapp_vserver", 'fake_vserver',
                          group=self.backend)
        data_motion.CONF.list_all_sections.return_value = [self.backend]

        config = data_motion.get_backend_configuration(self.backend)

        self.assertEqual('fake_vserver', config.netapp_vserver)

    def test_get_config_for_backend_different_backend_name(self):
        self.mock_object(data_motion, "CONF")
        CONF.set_override("netapp_vserver", 'fake_vserver',
                          group=self.backend)
        CONF.set_override("share_backend_name", "fake_backend_name",
                          group=self.backend)
        data_motion.CONF.list_all_sections.return_value = [self.backend]

        config = data_motion.get_backend_configuration(self.backend)

        self.assertEqual('fake_vserver', config.netapp_vserver)
        self.assertEqual('fake_backend_name', config.share_backend_name)

    @ddt.data([], ['fake_backend1', 'fake_backend2'])
    def test_get_config_for_backend_not_configured(self, conf_sections):
        self.mock_object(data_motion, "CONF")
        data_motion.CONF.list_all_sections.return_value = conf_sections

        self.assertRaises(exception.BadConfigurationException,
                          data_motion.get_backend_configuration,
                          self.backend)


@ddt.ddt
class NetAppCDOTDataMotionSessionTestCase(test.TestCase):

    def setUp(self):
        super(NetAppCDOTDataMotionSessionTestCase, self).setUp()
        self.source_backend = 'backend1'
        self.dest_backend = 'backend2'

        config = configuration.Configuration(driver.share_opts,
                                             config_group=self.source_backend)
        config.append_config_values(na_opts.netapp_cluster_opts)
        config.append_config_values(na_opts.netapp_connection_opts)
        config.append_config_values(na_opts.netapp_basicauth_opts)
        config.append_config_values(na_opts.netapp_transport_opts)
        config.append_config_values(na_opts.netapp_support_opts)
        config.append_config_values(na_opts.netapp_provisioning_opts)
        config.append_config_values(na_opts.netapp_data_motion_opts)

        self.mock_object(data_motion, "get_backend_configuration",
                         mock.Mock(return_value=config))

        self.mock_cmode_client = self.mock_object(client_cmode,
                                                  "NetAppCmodeClient",
                                                  mock.Mock())
        self.dm_session = data_motion.DataMotionSession()
        self.fake_src_share = copy.deepcopy(fake.SHARE)
        self.fake_src_share_server = copy.deepcopy(fake.SHARE_SERVER)
        self.source_vserver = 'source_vserver'
        self.source_backend_name = (
            self.fake_src_share_server['host'].split('@')[1])
        self.fake_src_share_server['backend_details']['vserver_name'] = (
            self.source_vserver
        )
        self.fake_src_share['share_server'] = self.fake_src_share_server
        self.fake_src_share['id'] = 'c02d497a-236c-4852-812a-0d39373e312a'
        self.fake_src_vol_name = 'share_c02d497a_236c_4852_812a_0d39373e312a'
        self.fake_dest_share = copy.deepcopy(fake.SHARE)
        self.fake_dest_share_server = copy.deepcopy(fake.SHARE_SERVER_2)
        self.dest_vserver = 'dest_vserver'
        self.dest_backend_name = (
            self.fake_dest_share_server['host'].split('@')[1])
        self.fake_dest_share_server['backend_details']['vserver_name'] = (
            self.dest_vserver
        )
        self.fake_dest_share['share_server'] = self.fake_dest_share_server
        self.fake_dest_share['id'] = '34fbaf57-745d-460f-8270-3378c2945e30'
        self.fake_dest_vol_name = 'share_34fbaf57_745d_460f_8270_3378c2945e30'

        self.mock_src_client = mock.Mock()
        self.mock_dest_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[self.mock_dest_client,
                                                self.mock_src_client]))
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(side_effect=[
                             (self.mock_src_client, self.source_vserver),
                             (self.mock_dest_client, self.dest_vserver)]))

    def test_get_client_and_vserver_name(self):
        dm_session = data_motion.DataMotionSession()
        client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(return_value=client))

        result = dm_session.get_client_and_vserver_name(fake.SHARE_SERVER)
        expected = (client,
                    fake.SHARE_SERVER['backend_details']['vserver_name'])

        self.assertEqual(expected, result)
        data_motion.get_client_for_backend.assert_called_once_with(
            fake.BACKEND_NAME, vserver_name=fake.VSERVER1
        )

    @ddt.data(True, False)
    def test_create_snapmirror_mount(self, mount):
        mock_dest_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(return_value=mock_dest_client))
        self.mock_object(self.dm_session, 'wait_for_mount_replica')
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_mount_replica_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.create_snapmirror(self.fake_src_share,
                                          self.fake_dest_share,
                                          'data_protection', mount=mount)

        mock_dest_client.create_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, 'data_protection', schedule='hourly'
        )
        mock_dest_client.initialize_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        if mount:
            self.dm_session.wait_for_mount_replica.assert_called_once_with(
                mock_dest_client, self.fake_dest_vol_name, timeout=30)
        else:
            self.dm_session.wait_for_mount_replica.assert_not_called()

    def test_create_snapmirror_svm(self):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))
        self.mock_object(self.dm_session, 'get_vserver_from_share_server',
                         mock.Mock(return_value=self.source_vserver))
        policy_name = 'policy_' + self.dest_vserver
        get_snapmirro_policy_name = self.mock_object(
            self.dm_session, '_get_backend_snapmirror_policy_name_svm',
            mock.Mock(return_value=policy_name))

        self.dm_session.create_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server)

        self.dm_session.get_client_and_vserver_name.assert_called_once_with(
            self.fake_dest_share_server
        )
        self.dm_session.get_vserver_from_share_server.assert_called_once_with(
            self.fake_src_share_server
        )
        get_snapmirro_policy_name.assert_called_once_with(
            self.fake_dest_share_server['id'], self.dest_backend_name
        )
        mock_dest_client.create_snapmirror_policy.assert_called_once_with(
            policy_name
        )
        mock_dest_client.create_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver,
            policy=policy_name, schedule='hourly'
        )
        mock_dest_client.initialize_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )

    def test_delete_snapmirror(self):
        mock_src_client = mock.Mock()
        mock_dest_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[mock_dest_client,
                                                mock_src_client]))
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))
        mock_wait_for_snapmirror_release_vol = self.mock_object(
            self.dm_session, 'wait_for_snapmirror_release_vol')

        self.dm_session.delete_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, clear_checkpoint=False
        )
        mock_dest_client.delete_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        mock_wait_for_snapmirror_release_vol.assert_called_once_with(
            self.source_vserver, self.dest_vserver, self.fake_src_vol_name,
            self.fake_dest_vol_name, False, mock_src_client,
            timeout=30
        )

    @ddt.data(True, False)
    def test_delete_snapmirror_svm(self, call_release):
        self.mock_object(self.dm_session, 'wait_for_snapmirror_release_svm')
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.delete_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server,
                                              release=call_release)

        self.mock_dest_client.abort_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        self.mock_dest_client.delete_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        if call_release:
            release_mock = self.dm_session.wait_for_snapmirror_release_svm
            release_mock.assert_called_once_with(
                self.source_vserver, self.dest_vserver, self.mock_src_client,
                timeout=mock_backend_config.netapp_snapmirror_release_timeout
            )

    def test_delete_snapmirror_does_not_exist(self):
        """Ensure delete succeeds when the snapmirror does not exist."""
        mock_src_client = mock.Mock()
        mock_dest_client = mock.Mock()
        mock_dest_client.abort_snapmirror_vol.side_effect = (
            netapp_api.NaApiError(code=netapp_api.EAPIERROR))
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[mock_dest_client,
                                                mock_src_client]))
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))
        mock_wait_for_snapmirror_release_vol = self.mock_object(
            self.dm_session, 'wait_for_snapmirror_release_vol')

        self.dm_session.delete_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, clear_checkpoint=False
        )
        mock_dest_client.delete_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        mock_wait_for_snapmirror_release_vol.assert_called_once_with(
            self.source_vserver, self.dest_vserver, self.fake_src_vol_name,
            self.fake_dest_vol_name, False, mock_src_client,
            timeout=30
        )

    def test_delete_snapmirror_svm_does_not_exist(self):
        """Ensure delete succeeds when the snapmirror does not exist."""
        self.mock_dest_client.abort_snapmirror_svm.side_effect = (
            netapp_api.NaApiError(code=netapp_api.EAPIERROR))
        self.mock_object(self.dm_session, 'wait_for_snapmirror_release_svm')
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.delete_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server)

        self.mock_dest_client.abort_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        self.mock_dest_client.delete_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        release_mock = self.dm_session.wait_for_snapmirror_release_svm
        release_mock.assert_called_once_with(
            self.source_vserver, self.dest_vserver, self.mock_src_client,
            timeout=mock_backend_config.netapp_snapmirror_release_timeout
        )

    def test_delete_snapmirror_error_deleting(self):
        """Ensure delete succeeds when the snapmirror does not exist."""
        mock_src_client = mock.Mock()
        mock_dest_client = mock.Mock()
        mock_dest_client.delete_snapmirror_vol.side_effect = (
            netapp_api.NaApiError(code=netapp_api.ESOURCE_IS_DIFFERENT))
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[mock_dest_client,
                                                mock_src_client]))
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))
        mock_wait_for_snapmirror_release_vol = self.mock_object(
            self.dm_session, 'wait_for_snapmirror_release_vol')

        self.dm_session.delete_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, clear_checkpoint=False
        )
        mock_dest_client.delete_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        mock_wait_for_snapmirror_release_vol.assert_called_once_with(
            self.source_vserver, self.dest_vserver, self.fake_src_vol_name,
            self.fake_dest_vol_name, False, mock_src_client,
            timeout=30
        )

    def test_delete_snapmirror_svm_error_deleting(self):
        """Ensure delete succeeds when the snapmirror does not exist."""
        self.mock_dest_client.delete_snapmirror_svm.side_effect = (
            netapp_api.NaApiError(code=netapp_api.ESOURCE_IS_DIFFERENT))
        self.mock_object(self.dm_session, 'wait_for_snapmirror_release_svm')
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_release_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.delete_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server)

        self.mock_dest_client.abort_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        self.mock_dest_client.delete_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver
        )
        release_mock = self.dm_session.wait_for_snapmirror_release_svm
        release_mock.assert_called_once_with(
            self.source_vserver, self.dest_vserver, self.mock_src_client,
            timeout=mock_backend_config.netapp_snapmirror_release_timeout
        )

    def test_delete_snapmirror_without_release(self):
        mock_src_client = mock.Mock()
        mock_dest_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[mock_dest_client,
                                                mock_src_client]))
        mock_wait_for_snapmirror_release_vol = self.mock_object(
            self.dm_session, 'wait_for_snapmirror_release_vol')

        self.dm_session.delete_snapmirror(self.fake_src_share,
                                          self.fake_dest_share,
                                          release=False)

        mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, clear_checkpoint=False
        )
        mock_dest_client.delete_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        self.assertFalse(mock_wait_for_snapmirror_release_vol.called)

    def test_delete_snapmirror_source_unreachable(self):
        mock_dest_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[mock_dest_client,
                                                Exception]))
        mock_wait_for_snapmirror_release_vol = self.mock_object(
            self.dm_session, 'wait_for_snapmirror_release_vol')

        self.dm_session.delete_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name, clear_checkpoint=False
        )
        mock_dest_client.delete_snapmirror_vol.assert_called_once_with(
            mock.ANY, self.fake_src_vol_name, mock.ANY,
            self.fake_dest_vol_name
        )
        self.assertFalse(mock_wait_for_snapmirror_release_vol.called)

    def test_break_snapmirror(self):
        self.mock_object(self.dm_session, 'quiesce_then_abort')

        self.dm_session.break_snapmirror(self.fake_src_share,
                                         self.fake_dest_share)

        self.mock_dest_client.break_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

        self.dm_session.quiesce_then_abort.assert_called_once_with(
            self.fake_src_share, self.fake_dest_share)

        self.mock_dest_client.mount_volume.assert_called_once_with(
            self.fake_dest_vol_name)

    def test_break_snapmirror_no_mount(self):
        self.mock_object(self.dm_session, 'quiesce_then_abort')

        self.dm_session.break_snapmirror(self.fake_src_share,
                                         self.fake_dest_share,
                                         mount=False)

        self.mock_dest_client.break_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

        self.dm_session.quiesce_then_abort.assert_called_once_with(
            self.fake_src_share, self.fake_dest_share)

        self.assertFalse(self.mock_dest_client.mount_volume.called)

    def test_break_snapmirror_wait_for_quiesced(self):
        self.mock_object(self.dm_session, 'quiesce_then_abort')

        self.dm_session.break_snapmirror(self.fake_src_share,
                                         self.fake_dest_share)

        self.dm_session.quiesce_then_abort.assert_called_once_with(
            self.fake_src_share, self.fake_dest_share)

        self.mock_dest_client.break_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

        self.mock_dest_client.mount_volume.assert_called_once_with(
            self.fake_dest_vol_name)

    def test_quiesce_then_abort_timeout(self):
        self.mock_object(time, 'sleep')
        mock_get_snapmirrors = mock.Mock(
            return_value=[{'relationship-status': "transferring"}])
        self.mock_object(self.mock_dest_client, 'get_snapmirrors',
                         mock_get_snapmirrors)
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_quiesce_timeout = 10
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.quiesce_then_abort(self.fake_src_share,
                                           self.fake_dest_share)

        self.mock_dest_client.get_snapmirrors.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            source_volume=self.fake_src_vol_name,
            dest_volume=self.fake_dest_vol_name,
            desired_attributes=['relationship-status', 'mirror-state']
        )
        self.assertEqual(2, self.mock_dest_client.get_snapmirrors.call_count)

        self.mock_dest_client.quiesce_snapmirror_vol.assert_called_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

        self.mock_dest_client.abort_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name,
            clear_checkpoint=False
        )

    def test_quiesce_then_abort_svm_timeout(self):
        self.mock_object(time, 'sleep')
        mock_get_snapmirrors = mock.Mock(
            return_value=[{'relationship-status': "transferring"}])
        self.mock_object(self.mock_dest_client, 'get_snapmirrors_svm',
                         mock_get_snapmirrors)
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_snapmirror_quiesce_timeout = 10
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))

        self.dm_session.quiesce_then_abort_svm(self.fake_src_share_server,
                                               self.fake_dest_share_server)

        self.mock_dest_client.get_snapmirrors_svm.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            desired_attributes=['relationship-status', 'mirror-state']
        )
        self.assertEqual(2,
                         self.mock_dest_client.get_snapmirrors_svm.call_count)

        self.mock_dest_client.quiesce_snapmirror_svm.assert_called_with(
            self.source_vserver, self.dest_vserver)

        self.mock_dest_client.abort_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver,
            clear_checkpoint=False
        )

    def test_quiesce_then_abort_wait_for_quiesced(self):
        self.mock_object(time, 'sleep')
        self.mock_object(self.mock_dest_client, 'get_snapmirrors',
                         mock.Mock(side_effect=[
                             [{'relationship-status': "transferring"}],
                             [{'relationship-status': "quiesced"}]]))

        self.dm_session.quiesce_then_abort(self.fake_src_share,
                                           self.fake_dest_share)

        self.mock_dest_client.get_snapmirrors.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            source_volume=self.fake_src_vol_name,
            dest_volume=self.fake_dest_vol_name,
            desired_attributes=['relationship-status', 'mirror-state']
        )
        self.assertEqual(2, self.mock_dest_client.get_snapmirrors.call_count)

        self.mock_dest_client.quiesce_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

    def test_quiesce_then_abort_svm_wait_for_quiesced(self):
        self.mock_object(time, 'sleep')
        self.mock_object(self.mock_dest_client, 'get_snapmirrors_svm',
                         mock.Mock(side_effect=[
                             [{'relationship-status': "transferring"}],
                             [{'relationship-status': "quiesced"}]]))

        self.dm_session.quiesce_then_abort_svm(self.fake_src_share_server,
                                               self.fake_dest_share_server)

        self.mock_dest_client.get_snapmirrors_svm.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            desired_attributes=['relationship-status', 'mirror-state']
        )
        self.assertEqual(2,
                         self.mock_dest_client.get_snapmirrors_svm.call_count)

        self.mock_dest_client.quiesce_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver)

    def test_resync_snapmirror(self):
        self.dm_session.resync_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        self.mock_dest_client.resync_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

    def test_change_snapmirror_source(self):
        fake_new_src_share = copy.deepcopy(fake.SHARE)
        fake_new_src_share['id'] = 'd02d497a-236c-4852-812a-0d39373e312a'
        fake_new_src_share_name = 'share_d02d497a_236c_4852_812a_0d39373e312a'
        mock_new_src_client = mock.Mock()
        self.mock_object(self.dm_session, 'delete_snapmirror')
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[self.mock_dest_client,
                                                self.mock_src_client,
                                                self.mock_dest_client,
                                                mock_new_src_client]))
        self.mock_object(na_utils, 'get_relationship_type',
                         mock.Mock(return_value=na_utils.DATA_PROTECTION_TYPE))

        self.dm_session.change_snapmirror_source(
            self.fake_dest_share, self.fake_src_share, fake_new_src_share,
            [self.fake_dest_share, self.fake_src_share, fake_new_src_share])

        self.assertFalse(self.mock_src_client.release_snapmirror_vol.called)

        self.assertEqual(4, self.dm_session.delete_snapmirror.call_count)
        self.dm_session.delete_snapmirror.assert_called_with(
            mock.ANY, mock.ANY, release=False, relationship_info_only=False
        )

        na_utils.get_relationship_type.assert_called_once_with(False)
        self.mock_dest_client.create_snapmirror_vol.assert_called_once_with(
            mock.ANY, fake_new_src_share_name, mock.ANY,
            self.fake_dest_vol_name, na_utils.DATA_PROTECTION_TYPE,
            schedule='hourly'
        )

        self.mock_dest_client.resync_snapmirror_vol.assert_called_once_with(
            mock.ANY, fake_new_src_share_name, mock.ANY,
            self.fake_dest_vol_name
        )

    def test_change_snapmirror_source_dhss_true(self):
        fake_new_src_share = copy.deepcopy(self.fake_src_share)
        fake_new_src_share['id'] = 'd02d497a-236c-4852-812a-0d39373e312a'
        fake_new_src_share_name = 'share_d02d497a_236c_4852_812a_0d39373e312a'
        fake_new_src_share_server = fake_new_src_share['share_server']
        fake_new_src_ss_name = (
            fake_new_src_share_server['backend_details']['vserver_name'])
        self.mock_object(self.dm_session, 'delete_snapmirror')
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(side_effect=[self.mock_dest_client,
                                                self.mock_src_client]))
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.driver_handles_share_servers = True
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))
        self.mock_object(self.mock_dest_client, 'get_vserver_peers',
                         mock.Mock(return_value=[]))
        peer_cluster_name = 'new_src_cluster_name'
        self.mock_object(self.mock_src_client, 'get_cluster_name',
                         mock.Mock(return_value=peer_cluster_name))
        self.mock_object(na_utils, 'get_relationship_type',
                         mock.Mock(return_value=na_utils.DATA_PROTECTION_TYPE))

        self.dm_session.change_snapmirror_source(
            self.fake_dest_share, self.fake_src_share, fake_new_src_share,
            [self.fake_dest_share, self.fake_src_share, fake_new_src_share])

        self.assertEqual(4, self.dm_session.delete_snapmirror.call_count)

        self.mock_dest_client.get_vserver_peers.assert_called_once_with(
            self.dest_vserver, fake_new_src_ss_name
        )
        self.assertTrue(self.mock_src_client.get_cluster_name.called)
        self.mock_dest_client.create_vserver_peer.assert_called_once_with(
            self.dest_vserver, fake_new_src_ss_name,
            peer_cluster_name=peer_cluster_name
        )
        self.mock_src_client.accept_vserver_peer.assert_called_once_with(
            fake_new_src_ss_name, self.dest_vserver
        )
        na_utils.get_relationship_type.assert_called_once_with(False)
        self.dm_session.delete_snapmirror.assert_called_with(
            mock.ANY, mock.ANY, release=False, relationship_info_only=False
        )
        self.mock_dest_client.create_snapmirror_vol.assert_called_once_with(
            mock.ANY, fake_new_src_share_name, mock.ANY,
            self.fake_dest_vol_name, na_utils.DATA_PROTECTION_TYPE,
            schedule='hourly'
        )
        self.mock_dest_client.resync_snapmirror_vol.assert_called_once_with(
            mock.ANY, fake_new_src_share_name, mock.ANY,
            self.fake_dest_vol_name
        )

    def test_get_snapmirrors(self):
        self.mock_object(self.mock_dest_client, 'get_snapmirrors')

        self.dm_session.get_snapmirrors(self.fake_src_share,
                                        self.fake_dest_share)

        self.mock_dest_client.get_snapmirrors.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            source_volume=self.fake_src_vol_name,
            dest_volume=self.fake_dest_vol_name,
            desired_attributes=['relationship-status',
                                'mirror-state',
                                'source-vserver',
                                'source-volume',
                                'last-transfer-end-timestamp']
        )
        self.assertEqual(1, self.mock_dest_client.get_snapmirrors.call_count)

    def test_get_snapmirrors_svm(self):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))
        self.mock_object(mock_dest_client, 'get_snapmirrors_svm')

        self.dm_session.get_snapmirrors_svm(self.fake_src_share_server,
                                            self.fake_dest_share_server)

        mock_dest_client.get_snapmirrors_svm.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
            desired_attributes=['relationship-status',
                                'mirror-state',
                                'last-transfer-end-timestamp']
        )
        self.assertEqual(1, mock_dest_client.get_snapmirrors_svm.call_count)

    def test_get_snapmirror_destinations_svm(self):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))
        self.mock_object(mock_dest_client, 'get_snapmirror_destinations_svm')

        self.dm_session.get_snapmirror_destinations_svm(
            self.fake_src_share_server, self.fake_dest_share_server)

        mock_dest_client.get_snapmirror_destinations_svm.assert_called_with(
            source_vserver=self.source_vserver,
            dest_vserver=self.dest_vserver,
        )
        self.assertEqual(1, mock_dest_client.get_snapmirror_destinations_svm
                         .call_count)

    def test_update_snapmirror(self):
        self.mock_object(self.mock_dest_client, 'get_snapmirrors')

        self.dm_session.update_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        self.mock_dest_client.update_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

    def test_update_snapmirror_svm(self):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))

        self.dm_session.update_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server)

        mock_dest_client.update_snapmirror_svm.assert_called_once_with(
            self.source_vserver, self.dest_vserver)

    def test_abort_and_break_snapmirror_svm(self):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))
        self.mock_object(self.dm_session, 'quiesce_then_abort_svm')

        self.dm_session.quiesce_and_break_snapmirror_svm(
            self.fake_src_share_server, self.fake_dest_share_server
        )

        self.dm_session.get_client_and_vserver_name.assert_called_once_with(
            self.fake_dest_share_server
        )
        self.dm_session.quiesce_then_abort_svm.assert_called_once_with(
            self.fake_src_share_server, self.fake_dest_share_server
        )
        mock_dest_client.break_snapmirror_svm(self.source_vserver,
                                              self.dest_vserver)

    @ddt.data({'snapmirrors': ['fake_snapmirror'],
               'vserver_subtype': 'default'},
              {'snapmirrors': [],
               'vserver_subtype': 'default'},
              {'snapmirrors': [],
               'vserver_subtype': 'dp_destination'})
    @ddt.unpack
    def test_cancel_snapmirror_svm(self, snapmirrors, vserver_subtype):
        mock_dest_client = mock.Mock()
        self.mock_object(self.dm_session, 'get_client_and_vserver_name',
                         mock.Mock(return_value=(mock_dest_client,
                                                 self.dest_vserver)))
        mock_backend_config = na_fakes.create_configuration()
        mock_backend_config.netapp_server_migration_state_change_timeout = 30
        self.mock_object(data_motion, 'get_backend_configuration',
                         mock.Mock(return_value=mock_backend_config))
        self.mock_object(self.dm_session, 'get_snapmirrors_svm',
                         mock.Mock(return_value=snapmirrors))
        self.mock_object(self.dm_session, 'quiesce_and_break_snapmirror_svm')
        self.mock_object(self.dm_session, 'wait_for_vserver_state')
        self.mock_object(self.dm_session, 'delete_snapmirror_svm')
        vserver_info = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info['subtype'] = vserver_subtype
        self.mock_object(mock_dest_client, 'get_vserver_info',
                         mock.Mock(return_value=vserver_info))
        self.mock_object(self.dm_session, 'convert_svm_to_default_subtype')

        self.dm_session.cancel_snapmirror_svm(self.fake_src_share_server,
                                              self.fake_dest_share_server)

        data_motion.get_backend_configuration.assert_called_once_with(
            self.dest_backend_name
        )
        self.dm_session.get_client_and_vserver_name.assert_called_once_with(
            self.fake_dest_share_server
        )
        self.dm_session.get_snapmirrors_svm.assert_called_once_with(
            self.fake_src_share_server, self.fake_dest_share_server
        )
        if snapmirrors:
            quiesce_mock = self.dm_session.quiesce_and_break_snapmirror_svm
            quiesce_mock.assert_called_once_with(
                self.fake_src_share_server, self.fake_dest_share_server
            )
            self.dm_session.wait_for_vserver_state.assert_called_once_with(
                self.dest_vserver, mock_dest_client, subtype='default',
                state='running', operational_state='stopped',
                timeout=(mock_backend_config
                         .netapp_server_migration_state_change_timeout)
            )
            self.dm_session.delete_snapmirror_svm.assert_called_once_with(
                self.fake_src_share_server, self.fake_dest_share_server
            )
        else:
            mock_dest_client.get_vserver_info.assert_called_once_with(
                self.dest_vserver
            )
            convert_svm = self.dm_session.convert_svm_to_default_subtype
            if vserver_subtype == 'dp_destination':
                convert_svm.assert_called_once_with(
                    self.dest_vserver, mock_dest_client,
                    timeout=(mock_backend_config
                             .netapp_server_migration_state_change_timeout)
                )
            else:
                self.assertFalse(convert_svm.called)

    def test_resume_snapmirror(self):
        self.mock_object(self.mock_dest_client, 'get_snapmirrors')

        self.dm_session.resume_snapmirror(self.fake_src_share,
                                          self.fake_dest_share)

        self.mock_dest_client.resume_snapmirror_vol.assert_called_once_with(
            self.source_vserver, self.fake_src_vol_name,
            self.dest_vserver, self.fake_dest_vol_name)

    @ddt.data((None, exception.StorageCommunicationException),
              (exception.StorageCommunicationException, None))
    @ddt.unpack
    def test_remove_qos_on_old_active_replica_unreachable_backend(self,
                                                                  side_eff_1,
                                                                  side_eff_2):
        mock_source_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(return_value=mock_source_client))
        self.mock_object(
            mock_source_client, 'set_qos_policy_group_for_volume',
            mock.Mock(side_effect=side_eff_1))
        self.mock_object(
            mock_source_client, 'mark_qos_policy_group_for_deletion',
            mock.Mock(side_effect=side_eff_2))
        self.mock_object(data_motion.LOG, 'exception')

        retval = self.dm_session.remove_qos_on_old_active_replica(
            self.fake_src_share)

        self.assertIsNone(retval)
        (mock_source_client.set_qos_policy_group_for_volume
         .assert_called_once_with(self.fake_src_vol_name, 'none'))
        data_motion.LOG.exception.assert_called_once()

    def test_remove_qos_on_old_active_replica(self):
        mock_source_client = mock.Mock()
        self.mock_object(data_motion, 'get_client_for_backend',
                         mock.Mock(return_value=mock_source_client))
        self.mock_object(data_motion.LOG, 'exception')

        retval = self.dm_session.remove_qos_on_old_active_replica(
            self.fake_src_share)

        self.assertIsNone(retval)
        (mock_source_client.set_qos_policy_group_for_volume
         .assert_called_once_with(self.fake_src_vol_name, 'none'))
        data_motion.LOG.exception.assert_not_called()

    @ddt.data(True, False)
    def test_convert_svm_to_default_subtype(self, is_dest):
        mock_client = mock.Mock()
        vserver_info_default = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_default['subtype'] = 'default'
        vserver_info_dp = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_dp['subtype'] = 'dp_destination'
        self.mock_object(mock_client, 'get_vserver_info',
                         mock.Mock(side_effect=[vserver_info_dp,
                                                vserver_info_default]))
        self.mock_object(mock_client, 'break_snapmirror_svm')

        self.dm_session.convert_svm_to_default_subtype(fake.VSERVER1,
                                                       mock_client,
                                                       is_dest_path=is_dest,
                                                       timeout=20)

        mock_client.get_vserver_info.assert_has_calls([
            mock.call(fake.VSERVER1), mock.call(fake.VSERVER1)])
        if is_dest:
            mock_client.break_snapmirror_svm.assert_called_once_with(
                dest_vserver=fake.VSERVER1
            )
        else:
            mock_client.break_snapmirror_svm.assert_called_once_with(
                source_vserver=fake.VSERVER1
            )

    def test_convert_svm_to_default_subtype_timeout(self):
        mock_client = mock.Mock()
        vserver_info_dp = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_dp['subtype'] = 'dp_destination'
        self.mock_object(mock_client, 'get_vserver_info',
                         mock.Mock(side_effect=[vserver_info_dp]))
        self.mock_object(mock_client, 'break_snapmirror_svm')

        self.assertRaises(
            exception.NetAppException,
            self.dm_session.convert_svm_to_default_subtype,
            fake.VSERVER1, mock_client, is_dest_path=True, timeout=10)

        mock_client.get_vserver_info.assert_called_once_with(fake.VSERVER1)
        mock_client.break_snapmirror_svm.assert_called_once_with(
            dest_vserver=fake.VSERVER1)

    def test_wait_for_vserver_state(self,):
        mock_client = mock.Mock()
        vserver_info_default = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_default['subtype'] = 'default'
        vserver_info_dp = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_dp['subtype'] = 'dp_destination'
        self.mock_object(mock_client, 'get_vserver_info',
                         mock.Mock(side_effect=[vserver_info_dp,
                                                vserver_info_default]))

        self.dm_session.wait_for_vserver_state(fake.VSERVER1, mock_client,
                                               state='running',
                                               operational_state='running',
                                               subtype='default', timeout=20)

        mock_client.get_vserver_info.assert_has_calls([
            mock.call(fake.VSERVER1), mock.call(fake.VSERVER1)])

    def test_wait_for_vserver_state_timeout(self):
        mock_client = mock.Mock()
        vserver_info_dp = copy.deepcopy(fake.VSERVER_INFO)
        vserver_info_dp['subtype'] = 'dp_destination'
        self.mock_object(mock_client, 'get_vserver_info',
                         mock.Mock(side_effect=[vserver_info_dp]))

        self.assertRaises(
            exception.NetAppException,
            self.dm_session.wait_for_vserver_state,
            fake.VSERVER1, mock_client, state='running',
            operational_state='running', subtype='default', timeout=10)

        mock_client.get_vserver_info.assert_called_once_with(fake.VSERVER1)

    @ddt.data(mock.Mock(),
              mock.Mock(side_effect=netapp_api.NaApiError(
                  code=netapp_api.EOBJECTNOTFOUND)))
    def test_wait_for_snapmirror_release_svm(self, release_snapmirror_ret):
        src_mock_client = mock.Mock()
        get_snapmirrors_mock = self.mock_object(
            src_mock_client, 'get_snapmirror_destinations_svm',
            mock.Mock(side_effect=[['fake_snapmirror'], []]))
        self.mock_object(src_mock_client, 'release_snapmirror_svm',
                         release_snapmirror_ret)

        self.dm_session.wait_for_snapmirror_release_svm(fake.VSERVER1,
                                                        fake.VSERVER2,
                                                        src_mock_client,
                                                        timeout=20)
        get_snapmirrors_mock.assert_has_calls([
            mock.call(source_vserver=fake.VSERVER1,
                      dest_vserver=fake.VSERVER2),
            mock.call(source_vserver=fake.VSERVER1,
                      dest_vserver=fake.VSERVER2)])
        src_mock_client.release_snapmirror_svm.assert_called_once_with(
            fake.VSERVER1, fake.VSERVER2)

    def test_wait_for_snapmirror_release_svm_timeout(self):
        src_mock_client = mock.Mock()
        get_snapmirrors_mock = self.mock_object(
            src_mock_client, 'get_snapmirror_destinations_svm',
            mock.Mock(side_effect=[['fake_snapmirror']]))
        self.mock_object(src_mock_client, 'release_snapmirror_svm')

        self.assertRaises(exception.NetAppException,
                          self.dm_session.wait_for_snapmirror_release_svm,
                          fake.VSERVER1, fake.VSERVER2,
                          src_mock_client, timeout=10)

        get_snapmirrors_mock.assert_called_once_with(
            source_vserver=fake.VSERVER1, dest_vserver=fake.VSERVER2)
        src_mock_client.release_snapmirror_svm.assert_called_once_with(
            fake.VSERVER1, fake.VSERVER2
        )

    def test_wait_for_mount_replica(self):

        mock_client = mock.Mock()
        self.mock_object(time, 'sleep')
        mock_warning_log = self.mock_object(data_motion.LOG, 'warning')

        self.dm_session.wait_for_mount_replica(
            mock_client, fake.SHARE_NAME)

        mock_client.mount_volume.ssert_called_once_with(fake.SHARE_NAME)
        self.assertEqual(0, mock_warning_log.call_count)

    def test_wait_for_mount_replica_timeout(self):

        mock_client = mock.Mock()
        self.mock_object(time, 'sleep')
        mock_warning_log = self.mock_object(data_motion.LOG, 'warning')
        undergoing_snapmirror = (
            'The volume is undergoing a snapmirror initialize.')
        na_api_error = netapp_api.NaApiError(code=netapp_api.EAPIERROR,
                                             message=undergoing_snapmirror)
        mock_client.mount_volume.side_effect = na_api_error

        self.assertRaises(exception.NetAppException,
                          self.dm_session.wait_for_mount_replica,
                          mock_client, fake.SHARE_NAME, timeout=30)

        self.assertEqual(3, mock_client.mount_volume.call_count)
        self.assertEqual(3, mock_warning_log.call_count)

    def test_wait_for_mount_replica_api_not_found(self):

        mock_client = mock.Mock()
        self.mock_object(time, 'sleep')
        mock_warning_log = self.mock_object(data_motion.LOG, 'warning')
        na_api_error = netapp_api.NaApiError(code=netapp_api.EOBJECTNOTFOUND)
        mock_client.mount_volume.side_effect = na_api_error

        self.assertRaises(exception.NetAppException,
                          self.dm_session.wait_for_mount_replica,
                          mock_client, fake.SHARE_NAME, timeout=30)

        mock_client.mount_volume.assert_called_once_with(fake.SHARE_NAME)
        mock_warning_log.assert_not_called()

    @ddt.data(mock.Mock(),
              mock.Mock(side_effect=netapp_api.NaApiError(
                  code=netapp_api.EOBJECTNOTFOUND)))
    def test_wait_for_snapmirror_release_vol(self, release_snapmirror_ret):
        src_mock_client = mock.Mock()
        get_snapmirrors_mock = self.mock_object(
            src_mock_client, 'get_snapmirror_destinations',
            mock.Mock(side_effect=[['fake_snapmirror'], []]))
        self.mock_object(src_mock_client, 'release_snapmirror_vol',
                         release_snapmirror_ret)

        self.dm_session.wait_for_snapmirror_release_vol(fake.VSERVER1,
                                                        fake.VSERVER2,
                                                        fake.SHARE_NAME,
                                                        fake.SHARE_NAME2,
                                                        False,
                                                        src_mock_client,
                                                        timeout=20)
        get_snapmirrors_mock.assert_has_calls([
            mock.call(source_vserver=fake.VSERVER1,
                      dest_vserver=fake.VSERVER2,
                      source_volume=fake.SHARE_NAME,
                      dest_volume=fake.SHARE_NAME2),
            mock.call(source_vserver=fake.VSERVER1,
                      dest_vserver=fake.VSERVER2,
                      source_volume=fake.SHARE_NAME,
                      dest_volume=fake.SHARE_NAME2)])
        src_mock_client.release_snapmirror_vol.assert_called_once_with(
            fake.VSERVER1, fake.SHARE_NAME, fake.VSERVER2, fake.SHARE_NAME2,
            relationship_info_only=False)

    def test_wait_for_snapmirror_release_vol_timeout(self):
        src_mock_client = mock.Mock()
        get_snapmirrors_mock = self.mock_object(
            src_mock_client, 'get_snapmirror_destinations',
            mock.Mock(side_effect=[['fake_snapmirror']]))
        self.mock_object(src_mock_client, 'release_snapmirror_vol')

        self.assertRaises(exception.NetAppException,
                          self.dm_session.wait_for_snapmirror_release_vol,
                          fake.VSERVER1, fake.VSERVER2, fake.SHARE_NAME,
                          fake.SHARE_NAME2, False, src_mock_client,
                          timeout=10)

        get_snapmirrors_mock.assert_has_calls([
            mock.call(source_vserver=fake.VSERVER1,
                      dest_vserver=fake.VSERVER2,
                      source_volume=fake.SHARE_NAME,
                      dest_volume=fake.SHARE_NAME2)])
        src_mock_client.release_snapmirror_vol.assert_called_once_with(
            fake.VSERVER1, fake.SHARE_NAME, fake.VSERVER2, fake.SHARE_NAME2,
            relationship_info_only=False)



# Copyright (c) 2019 Infortrend Technology, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log

from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers.infortrend import infortrend_nas

LOG = log.getLogger(__name__)

infortrend_nas_opts = [
    cfg.HostAddressOpt('infortrend_nas_ip',
                       required=True,
                       help='Infortrend NAS IP for management.'),
    cfg.StrOpt('infortrend_nas_user',
               default='manila',
               help='User for the Infortrend NAS server.'),
    cfg.StrOpt('infortrend_nas_password',
               default=None,
               secret=True,
               help='Password for the Infortrend NAS server. '
               'This is not necessary '
               'if infortrend_nas_ssh_key is set.'),
    cfg.StrOpt('infortrend_nas_ssh_key',
               default=None,
               help='SSH key for the Infortrend NAS server. '
               'This is not necessary '
               'if infortrend_nas_password is set.'),
    cfg.ListOpt('infortrend_share_pools',
                required=True,
                help='Comma separated list of Infortrend NAS pools.'),
    cfg.ListOpt('infortrend_share_channels',
                required=True,
                help='Comma separated list of Infortrend channels.'),
    cfg.IntOpt('infortrend_ssh_timeout',
               default=30,
               help='SSH timeout in seconds.'),
]

CONF = cfg.CONF
CONF.register_opts(infortrend_nas_opts)


class InfortrendNASDriver(driver.ShareDriver):

    """Infortrend Share Driver for GS/GSe Family using NASCLI.

    Version history:
        1.0.0 - Initial driver
    """

    VERSION = "1.0.0"
    PROTOCOL = "NFS_CIFS"

    def __init__(self, *args, **kwargs):
        super(InfortrendNASDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(infortrend_nas_opts)

        nas_ip = self.configuration.safe_get('infortrend_nas_ip')
        username = self.configuration.safe_get('infortrend_nas_user')
        password = self.configuration.safe_get('infortrend_nas_password')
        ssh_key = self.configuration.safe_get('infortrend_nas_ssh_key')
        timeout = self.configuration.safe_get('infortrend_ssh_timeout')
        self.backend_name = self.configuration.safe_get('share_backend_name')

        if not (password or ssh_key):
            msg = _('Either infortrend_nas_password or infortrend_nas_ssh_key '
                    'should be set.')
            raise exception.InvalidParameterValue(err=msg)

        pool_dict = self._init_pool_dict()
        channel_dict = self._init_channel_dict()
        self.ift_nas = infortrend_nas.InfortrendNAS(nas_ip, username, password,
                                                    ssh_key, timeout,
                                                    pool_dict, channel_dict)

    def _init_pool_dict(self):
        pools_names = self.configuration.safe_get('infortrend_share_pools')

        return {el: {} for el in pools_names}

    def _init_channel_dict(self):
        channels = self.configuration.safe_get('infortrend_share_channels')

        return {el: '' for el in channels}

    def do_setup(self, context):
        """Any initialization the share driver does while starting."""
        LOG.debug('Infortrend NAS do_setup start.')
        self.ift_nas.do_setup()

    def check_for_setup_error(self):
        """Check for setup error."""
        LOG.debug('Infortrend NAS check_for_setup_error start.')
        self.ift_nas.check_for_setup_error()

    def _update_share_stats(self):
        """Retrieve stats info from share group."""

        LOG.debug('Updating Infortrend backend [%s].', self.backend_name)

        data = dict(
            share_backend_name=self.backend_name,
            vendor_name='Infortrend',
            driver_version=self.VERSION,
            storage_protocol=self.PROTOCOL,
            reserved_percentage=self.configuration.reserved_share_percentage,
            reserved_snapshot_percentage=(
                self.configuration.reserved_share_from_snapshot_percentage
                or self.configuration.reserved_share_percentage),
            pools=self.ift_nas.update_pools_stats())
        LOG.debug('Infortrend pools status: %s', data['pools'])

        super(InfortrendNASDriver, self)._update_share_stats(data)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share.

        :param context: Current context
        :param share: Share model with share data.
        :param access_rules: All access rules for given share
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: Not used by this driver.

        :returns: None, or a dictionary of ``access_id``, ``access_key`` as
                  key: value pairs for the rules added, where, ``access_id``
                  is the UUID (string) of the access rule, and ``access_key``
                  is the credential (string) of the entity granted access.
                  During recovery after error, the returned dictionary must
                  contain ``access_id``, ``access_key`` for all the rules that
                  the driver is ordered to resync, i.e. rules in the
                  ``access_rules`` parameter.
        """

        return self.ift_nas.update_access(share, access_rules, add_rules,
                                          delete_rules, share_server)

    def create_share(self, context, share, share_server=None):
        """Create a share."""

        LOG.debug('Creating share: %s.', share['id'])

        return self.ift_nas.create_share(share, share_server)

    def delete_share(self, context, share, share_server=None):
        """Remove a share."""

        LOG.debug('Deleting share: %s.', share['id'])

        return self.ift_nas.delete_share(share, share_server)

    def get_pool(self, share):
        """Return pool name where the share resides on.

        :param share: The share hosted by the driver.
        """
        return self.ift_nas.get_pool(share)

    def ensure_share(self, context, share, share_server=None):
        """Invoked to ensure that share is exported.

        Driver can use this method to update the list of export locations of
        the share if it changes. To do that, you should return list with
        export locations.

        :return None or list with export locations
        """
        return self.ift_nas.ensure_share(share, share_server)

    def manage_existing(self, share, driver_options):
        """Brings an existing share under Manila management.

        If the provided share is not valid, then raise a
        ManageInvalidShare exception, specifying a reason for the failure.

        If the provided share is not in a state that can be managed, such as
        being replicated on the backend, the driver *MUST* raise
        ManageInvalidShare exception with an appropriate message.

        The share has a share_type, and the driver can inspect that and
        compare against the properties of the referenced backend share.
        If they are incompatible, raise a
        ManageExistingShareTypeMismatch, specifying a reason for the failure.

        :param share: Share model
        :param driver_options: Driver-specific options provided by admin.
        :return: share_update dictionary with required key 'size',
                 which should contain size of the share.
        """
        LOG.debug(
            'Manage existing for share: %(share)s,', {
                'share': share['share_id'],
            })
        return self.ift_nas.manage_existing(share, driver_options)

    def unmanage(self, share):
        """Removes the specified share from Manila management.

        Does not delete the underlying backend share.

        For most drivers, this will not need to do anything.  However, some
        drivers might use this call as an opportunity to clean up any
        Manila-specific configuration that they have associated with the
        backend share.

        If provided share cannot be unmanaged, then raise an
        UnmanageInvalidShare exception, specifying a reason for the failure.

        This method is invoked when the share is being unmanaged with
        a share type that has ``driver_handles_share_servers``
        extra-spec set to False.
        """
        LOG.debug(
            'Unmanage share: %(share)s', {
                'share': share['share_id'],
            })
        return self.ift_nas.unmanage(share)

    def extend_share(self, share, new_size, share_server=None):
        """Extends size of existing share.

        :param share: Share model
        :param new_size: New size of share (new_size > share['size'])
        :param share_server: Optional -- Share server model
        """
        return self.ift_nas.extend_share(share, new_size, share_server)

    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks size of existing share.

        If consumed space on share larger than new_size driver should raise
        ShareShrinkingPossibleDataLoss exception:
        raise ShareShrinkingPossibleDataLoss(share_id=share['id'])

        :param share: Share model
        :param new_size: New size of share (new_size < share['size'])
        :param share_server: Optional -- Share server model

        :raises ShareShrinkingPossibleDataLoss, NotImplementedError
        """
        return self.ift_nas.shrink_share(share, new_size, share_server)



# Copyright (c) 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

import ddt
from oslo_concurrency import processutils
from oslo_config import cfg
import six

from manila import exception
from manila.share import configuration
from manila.share.drivers import service_instance as generic_service_instance
from manila.share.drivers.windows import service_instance
from manila.share.drivers.windows import windows_utils
from manila import test

CONF = cfg.CONF
CONF.import_opt('driver_handles_share_servers',
                'manila.share.driver')
CONF.register_opts(generic_service_instance.common_opts)

serv_mgr_cls = service_instance.WindowsServiceInstanceManager
generic_serv_mgr_cls = generic_service_instance.ServiceInstanceManager


@ddt.ddt
class WindowsServiceInstanceManagerTestCase(test.TestCase):
    _FAKE_SERVER = {'ip': mock.sentinel.ip,
                    'instance_id': mock.sentinel.instance_id}

    @mock.patch.object(windows_utils, 'WindowsUtils')
    @mock.patch.object(serv_mgr_cls, '_check_auth_mode')
    def setUp(self, mock_check_auth, mock_utils_cls):
        self.flags(service_instance_user=mock.sentinel.username)
        self._remote_execute = mock.Mock()

        fake_conf = configuration.Configuration(None)
        self._mgr = serv_mgr_cls(remote_execute=self._remote_execute,
                                 driver_config=fake_conf)
        self._windows_utils = mock_utils_cls.return_value
        super(WindowsServiceInstanceManagerTestCase, self).setUp()

    @ddt.data({},
              {'use_cert_auth': False},
              {'use_cert_auth': False, 'valid_pass_complexity': False},
              {'certs_exist': False})
    @mock.patch('os.path.exists')
    @mock.patch.object(serv_mgr_cls, '_check_password_complexity')
    @ddt.unpack
    def test_check_auth_mode(self, mock_check_complexity, mock_path_exists,
                             use_cert_auth=True, certs_exist=True,
                             valid_pass_complexity=True):
        self.flags(service_instance_password=mock.sentinel.password)
        self._mgr._cert_pem_path = mock.sentinel.cert_path
        self._mgr._cert_key_pem_path = mock.sentinel.key_path
        mock_path_exists.return_value = certs_exist
        mock_check_complexity.return_value = valid_pass_complexity

        self._mgr._use_cert_auth = use_cert_auth

        invalid_auth = ((use_cert_auth and not certs_exist)
                        or not valid_pass_complexity)

        if invalid_auth:
            self.assertRaises(exception.ServiceInstanceException,
                              self._mgr._check_auth_mode)
        else:
            self._mgr._check_auth_mode()

        if not use_cert_auth:
            mock_check_complexity.assert_called_once_with(
                six.text_type(mock.sentinel.password))

    @ddt.data(False, True)
    def test_get_auth_info(self, use_cert_auth):
        self._mgr._use_cert_auth = use_cert_auth
        self._mgr._cert_pem_path = mock.sentinel.cert_path
        self._mgr._cert_key_pem_path = mock.sentinel.key_path

        auth_info = self._mgr._get_auth_info()

        expected_auth_info = {'use_cert_auth': use_cert_auth}
        if use_cert_auth:
            expected_auth_info.update(cert_pem_path=mock.sentinel.cert_path,
                                      cert_key_pem_path=mock.sentinel.key_path)

        self.assertEqual(expected_auth_info, auth_info)

    @mock.patch.object(serv_mgr_cls, '_get_auth_info')
    @mock.patch.object(generic_serv_mgr_cls, 'get_common_server')
    def test_common_server(self, mock_generic_get_server, mock_get_auth):
        mock_server_details = {'backend_details': {}}
        mock_auth_info = {'fake_auth_info': mock.sentinel.auth_info}

        mock_generic_get_server.return_value = mock_server_details
        mock_get_auth.return_value = mock_auth_info
        expected_server_details = dict(backend_details=mock_auth_info)

        server_details = self._mgr.get_common_server()

        mock_generic_get_server.assert_called_once_with()
        self.assertEqual(expected_server_details, server_details)

    @mock.patch.object(serv_mgr_cls, '_get_auth_info')
    @mock.patch.object(generic_serv_mgr_cls, '_get_new_instance_details')
    def test_get_new_instance_details(self, mock_generic_get_details,
                                      mock_get_auth):
        mock_server_details = {'fake_server_details':
                               mock.sentinel.server_details}
        mock_generic_get_details.return_value = mock_server_details
        mock_auth_info = {'fake_auth_info': mock.sentinel.auth_info}
        mock_get_auth.return_value = mock_auth_info

        expected_server_details = dict(mock_server_details, **mock_auth_info)
        instance_details = self._mgr._get_new_instance_details(
            server=mock.sentinel.server)

        mock_generic_get_details.assert_called_once_with(mock.sentinel.server)
        self.assertEqual(expected_server_details, instance_details)

    @ddt.data(('abAB01', True),
              ('abcdef', False),
              ('aA0', False))
    @ddt.unpack
    def test_check_password_complexity(self, password, expected_result):
        valid_complexity = self._mgr._check_password_complexity(
            password)
        self.assertEqual(expected_result, valid_complexity)

    @ddt.data(None, Exception)
    def test_server_connection(self, side_effect):
        self._remote_execute.side_effect = side_effect

        expected_result = side_effect is None
        is_available = self._mgr._test_server_connection(self._FAKE_SERVER)

        self.assertEqual(expected_result, is_available)
        self._remote_execute.assert_called_once_with(self._FAKE_SERVER,
                                                     "whoami",
                                                     retry=False)

    @ddt.data(False, True)
    def test_get_service_instance_create_kwargs(self, use_cert_auth):
        self._mgr._use_cert_auth = use_cert_auth
        self.flags(service_instance_password=mock.sentinel.admin_pass)

        if use_cert_auth:
            mock_cert_data = 'mock_cert_data'
            self.mock_object(service_instance, 'open',
                             mock.mock_open(
                                 read_data=mock_cert_data))
            expected_kwargs = dict(user_data=mock_cert_data)
        else:
            expected_kwargs = dict(
                meta=dict(admin_pass=six.text_type(mock.sentinel.admin_pass)))

        create_kwargs = self._mgr._get_service_instance_create_kwargs()

        self.assertEqual(expected_kwargs, create_kwargs)

    @mock.patch.object(generic_serv_mgr_cls, 'set_up_service_instance')
    @mock.patch.object(serv_mgr_cls, 'get_valid_security_service')
    @mock.patch.object(serv_mgr_cls, '_setup_security_service')
    def test_set_up_service_instance(self, mock_setup_security_service,
                                     mock_get_valid_security_service,
                                     mock_generic_setup_serv_inst):
        mock_service_instance = {'instance_details': None}
        mock_network_info = {'security_services':
                             mock.sentinel.security_services}

        mock_generic_setup_serv_inst.return_value = mock_service_instance
        mock_get_valid_security_service.return_value = (
            mock.sentinel.security_service)

        instance_details = self._mgr.set_up_service_instance(
            mock.sentinel.context, mock_network_info)

        mock_generic_setup_serv_inst.assert_called_once_with(
            mock.sentinel.context, mock_network_info)
        mock_get_valid_security_service.assert_called_once_with(
            mock.sentinel.security_services)

        mock_setup_security_service.assert_called_once_with(
            mock_service_instance, mock.sentinel.security_service)

        expected_instance_details = dict(mock_service_instance,
                                         joined_domain=True)
        self.assertEqual(expected_instance_details,
                         instance_details)

    @mock.patch.object(serv_mgr_cls, '_run_cloudbase_init_plugin_after_reboot')
    @mock.patch.object(serv_mgr_cls, '_join_domain')
    def test_setup_security_service(self, mock_join_domain,
                                    mock_run_cbsinit_plugin):
        utils = self._windows_utils
        mock_security_service = {'domain': mock.sentinel.domain,
                                 'user': mock.sentinel.admin_username,
                                 'password': mock.sentinel.admin_password,
                                 'dns_ip': mock.sentinel.dns_ip}
        utils.get_interface_index_by_ip.return_value = (
            mock.sentinel.interface_index)

        self._mgr._setup_security_service(self._FAKE_SERVER,
                                          mock_security_service)

        utils.set_dns_client_search_list.assert_called_once_with(
            self._FAKE_SERVER,
            [mock_security_service['domain']])
        utils.get_interface_index_by_ip.assert_called_once_with(
            self._FAKE_SERVER,
            self._FAKE_SERVER['ip'])
        utils.set_dns_client_server_addresses.assert_called_once_with(
            self._FAKE_SERVER,
            mock.sentinel.interface_index,
            [mock_security_service['dns_ip']])
        mock_run_cbsinit_plugin.assert_called_once_with(
            self._FAKE_SERVER,
            plugin_name=self._mgr._CBS_INIT_WINRM_PLUGIN)
        mock_join_domain.assert_called_once_with(
            self._FAKE_SERVER,
            mock.sentinel.domain,
            mock.sentinel.admin_username,
            mock.sentinel.admin_password)

    @ddt.data({'join_domain_side_eff': Exception},
              {'server_available': False,
               'expected_exception': exception.ServiceInstanceException},
              {'join_domain_side_eff': processutils.ProcessExecutionError,
               'expected_exception': processutils.ProcessExecutionError},
              {'domain_mismatch': True,
               'expected_exception': exception.ServiceInstanceException})
    @mock.patch.object(generic_serv_mgr_cls, 'reboot_server')
    @mock.patch.object(generic_serv_mgr_cls, 'wait_for_instance_to_be_active')
    @mock.patch.object(generic_serv_mgr_cls, '_check_server_availability')
    @ddt.unpack
    def test_join_domain(self, mock_check_avail,
                         mock_wait_instance_active,
                         mock_reboot_server,
                         expected_exception=None,
                         server_available=True,
                         domain_mismatch=False,
                         join_domain_side_eff=None):
        self._windows_utils.join_domain.side_effect = join_domain_side_eff
        mock_check_avail.return_value = server_available
        self._windows_utils.get_current_domain.return_value = (
            None if domain_mismatch else mock.sentinel.domain)
        domain_params = (mock.sentinel.domain,
                         mock.sentinel.admin_username,
                         mock.sentinel.admin_password)

        if expected_exception:
            self.assertRaises(expected_exception,
                              self._mgr._join_domain,
                              self._FAKE_SERVER,
                              *domain_params)
        else:
            self._mgr._join_domain(self._FAKE_SERVER,
                                   *domain_params)

        if join_domain_side_eff != processutils.ProcessExecutionError:
            mock_reboot_server.assert_called_once_with(
                self._FAKE_SERVER, soft_reboot=True)
            mock_wait_instance_active.assert_called_once_with(
                self._FAKE_SERVER['instance_id'],
                timeout=self._mgr.max_time_to_build_instance)
            mock_check_avail.assert_called_once_with(self._FAKE_SERVER)
            if server_available:
                self._windows_utils.get_current_domain.assert_called_once_with(
                    self._FAKE_SERVER)

        self._windows_utils.join_domain.assert_called_once_with(
            self._FAKE_SERVER,
            *domain_params)

    @ddt.data([],
              [{'type': 'active_directory'}],
              [{'type': 'active_directory'}] * 2,
              [{'type': mock.sentinel.invalid_type}])
    def test_get_valid_security_service(self, security_services):
        valid_security_service = self._mgr.get_valid_security_service(
            security_services)

        if (security_services and len(security_services) == 1 and
                security_services[0]['type'] == 'active_directory'):
            expected_valid_sec_service = security_services[0]
        else:
            expected_valid_sec_service = None

        self.assertEqual(expected_valid_sec_service,
                         valid_security_service)

    @mock.patch.object(serv_mgr_cls, '_get_cbs_init_reg_section')
    def test_run_cloudbase_init_plugin_after_reboot(self,
                                                    mock_get_cbs_init_reg):
        self._FAKE_SERVER = {'instance_id': mock.sentinel.instance_id}
        mock_get_cbs_init_reg.return_value = mock.sentinel.cbs_init_reg_sect
        expected_plugin_key_path = "%(cbs_init)s\\%(instance_id)s\\Plugins" % {
            'cbs_init': mock.sentinel.cbs_init_reg_sect,
            'instance_id': self._FAKE_SERVER['instance_id']}

        self._mgr._run_cloudbase_init_plugin_after_reboot(
            server=self._FAKE_SERVER,
            plugin_name=mock.sentinel.plugin_name)

        mock_get_cbs_init_reg.assert_called_once_with(self._FAKE_SERVER)
        self._windows_utils.set_win_reg_value.assert_called_once_with(
            self._FAKE_SERVER,
            path=expected_plugin_key_path,
            key=mock.sentinel.plugin_name,
            value=self._mgr._CBS_INIT_RUN_PLUGIN_AFTER_REBOOT)

    @ddt.data(
        {},
        {'exec_errors': [
            processutils.ProcessExecutionError(stderr='Cannot find path'),
            processutils.ProcessExecutionError(stderr='Cannot find path')],
         'expected_exception': exception.ServiceInstanceException},
        {'exec_errors': [processutils.ProcessExecutionError(stderr='')],
         'expected_exception': processutils.ProcessExecutionError},
        {'exec_errors': [
            processutils.ProcessExecutionError(stderr='Cannot find path'),
            None]}
    )
    @ddt.unpack
    def test_get_cbs_init_reg_section(self, exec_errors=None,
                                      expected_exception=None):
        self._windows_utils.normalize_path.return_value = (
            mock.sentinel.normalized_section_path)
        self._windows_utils.get_win_reg_value.side_effect = exec_errors

        if expected_exception:
            self.assertRaises(expected_exception,
                              self._mgr._get_cbs_init_reg_section,
                              mock.sentinel.server)
        else:
            cbs_init_section = self._mgr._get_cbs_init_reg_section(
                mock.sentinel.server)
            self.assertEqual(mock.sentinel.normalized_section_path,
                             cbs_init_section)

        base_path = 'hklm:\\SOFTWARE'
        cbs_section = 'Cloudbase Solutions\\Cloudbase-Init'
        tested_upper_sections = ['']

        if exec_errors and 'Cannot find path' in exec_errors[0].stderr:
            tested_upper_sections.append('Wow6432Node')

        tested_sections = [os.path.join(base_path,
                                        upper_section,
                                        cbs_section)
                           for upper_section in tested_upper_sections]
        self._windows_utils.normalize_path.assert_has_calls(
            [mock.call(tested_section)
             for tested_section in tested_sections])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Change the datetime precision for all objects (MySQL and derivatives)

Revision ID: fbdfabcba377
Revises: 478c445d8d3e
Create Date: 2020-04-20 15:32:59.365323

"""

# revision identifiers, used by Alembic.
revision = 'fbdfabcba377'
down_revision = '478c445d8d3e'

from alembic import op
from sqlalchemy import DateTime, dialects


# DB Tables that can be affected by low precision timestamps in MySQL:
TABLES = ('services', 'quotas', 'project_user_quotas', 'backend_info',
          'project_share_type_quotas', 'quota_classes', 'quota_usages',
          'reservations', 'shares', 'share_instance_export_locations',
          'share_instances', 'share_instance_export_locations_metadata',
          'share_types', 'share_type_projects', 'share_type_extra_specs',
          'share_metadata', 'share_access_map', 'share_access_rules_metadata',
          'share_instance_access_map', 'share_snapshot_instances',
          'share_snapshots', 'share_snapshot_access_map', 'share_networks',
          'share_snapshot_instance_access_map', 'share_network_subnets',
          'share_snapshot_instance_export_locations', 'security_services',
          'share_servers', 'share_server_backend_details', 'share_groups',
          'network_allocations', 'share_network_security_service_association',
          'drivers_private_data', 'availability_zones', 'share_group_types',
          'share_group_type_projects', 'share_group_type_specs', 'messages',
          'share_group_snapshots', 'share_group_type_share_type_mappings',
          'share_group_share_type_mappings')


def upgrade():
    context = op.get_context()

    if context.bind.dialect.name == 'mysql':
        # override the default precision of DateTime:
        for table in TABLES:
            op.alter_column(table, 'created_at',
                            existing_type=DateTime,
                            type_=dialects.mysql.DATETIME(fsp=6))
            op.alter_column(table, 'updated_at',
                            existing_type=DateTime,
                            type_=dialects.mysql.DATETIME(fsp=6))
            op.alter_column(table, 'deleted_at',
                            existing_type=DateTime,
                            type_=dialects.mysql.DATETIME(fsp=6))


def downgrade():
    context = op.get_context()

    if context.bind.dialect.name == 'mysql':
        for table in TABLES:
            op.alter_column(table, 'created_at',
                            existing_type=dialects.mysql.DATETIME(fsp=6),
                            type_=DateTime)
            op.alter_column(table, 'updated_at',
                            existing_type=dialects.mysql.DATETIME(fsp=6),
                            type_=DateTime)
            op.alter_column(table, 'deleted_at',
                            existing_type=dialects.mysql.DATETIME(fsp=6),
                            type_=DateTime)



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import errno
import os
import re

from oslo_config import cfg
from oslo_log import log

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share.drivers.ganesha import manager as ganesha_manager
from manila.share.drivers.ganesha import utils as ganesha_utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class NASHelperBase(object, metaclass=abc.ABCMeta):
    """Interface to work with share."""

    # drivers that use a helper derived from this class
    # should pass the following attributes to
    # ganesha_utils.validate_access_rule in their
    # update_access implementation.
    supported_access_types = ()
    supported_access_levels = ()

    def __init__(self, execute, config, **kwargs):
        self.configuration = config
        self._execute = execute

    def init_helper(self):
        """Initializes protocol-specific NAS drivers."""

    @abc.abstractmethod
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules of share."""


class GaneshaNASHelper(NASHelperBase):
    """Perform share access changes using Ganesha version < 2.4."""

    supported_access_types = ('ip', )
    supported_access_levels = (constants.ACCESS_LEVEL_RW,
                               constants.ACCESS_LEVEL_RO)

    def __init__(self, execute, config, tag='<no name>', **kwargs):
        super(GaneshaNASHelper, self).__init__(execute, config, **kwargs)
        self.tag = tag

    _confrx = re.compile(r'\.(conf|json)\Z')

    def _load_conf_dir(self, dirpath, must_exist=True):
        """Load Ganesha config files in dirpath in alphabetic order."""
        try:
            dirlist = os.listdir(dirpath)
        except OSError as e:
            if e.errno != errno.ENOENT or must_exist:
                raise
            dirlist = []
        LOG.info('Loading Ganesha config from %s.', dirpath)
        conf_files = list(filter(self._confrx.search, dirlist))
        conf_files.sort()
        export_template = {}
        for conf_file in conf_files:
            with open(os.path.join(dirpath, conf_file)) as f:
                ganesha_utils.patch(
                    export_template,
                    ganesha_manager.parseconf(f.read()))
        return export_template

    def init_helper(self):
        """Initializes protocol-specific NAS drivers."""
        self.ganesha = ganesha_manager.GaneshaManager(
            self._execute,
            self.tag,
            ganesha_config_path=self.configuration.ganesha_config_path,
            ganesha_export_dir=self.configuration.ganesha_export_dir,
            ganesha_db_path=self.configuration.ganesha_db_path,
            ganesha_service_name=self.configuration.ganesha_service_name)
        system_export_template = self._load_conf_dir(
            self.configuration.ganesha_export_template_dir,
            must_exist=False)
        if system_export_template:
            self.export_template = system_export_template
        else:
            self.export_template = self._default_config_hook()

    def _default_config_hook(self):
        """The default export block.

        Subclass this to add FSAL specific defaults.

        Suggested approach: take the return value of superclass'
        method, patch with dict containing your defaults, and
        return the result. However, you can also provide your
        defaults from scratch with no regard to superclass.
        """

        return self._load_conf_dir(ganesha_utils.path_from(__file__, "conf"))

    def _fsal_hook(self, base_path, share, access):
        """Subclass this to create FSAL block."""
        return {}

    def _cleanup_fsal_hook(self, base_path, share, access):
        """Callback for FSAL specific cleanup after removing an export."""
        pass

    def _allow_access(self, base_path, share, access):
        """Allow access to the share."""
        ganesha_utils.validate_access_rule(
            self.supported_access_types, self.supported_access_levels,
            access, abort=True)

        access = ganesha_utils.fixup_access_rule(access)

        cf = {}
        accid = access['id']
        name = share['name']
        export_name = "%s--%s" % (name, accid)
        ganesha_utils.patch(cf, self.export_template, {
            'EXPORT': {
                'Export_Id': self.ganesha.get_export_id(),
                'Path': os.path.join(base_path, name),
                'Pseudo': os.path.join(base_path, export_name),
                'Tag': accid,
                'CLIENT': {
                    'Clients': access['access_to']
                },
                'FSAL': self._fsal_hook(base_path, share, access)
            }
        })
        self.ganesha.add_export(export_name, cf)

    def _deny_access(self, base_path, share, access):
        """Deny access to the share."""
        self.ganesha.remove_export("%s--%s" % (share['name'], access['id']))

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules of share."""
        rule_state_map = {}
        if not (add_rules or delete_rules):
            add_rules = access_rules
            self.ganesha.reset_exports()
            self.ganesha.restart_service()

        for rule in add_rules:
            try:
                self._allow_access('/', share, rule)
            except (exception.InvalidShareAccess,
                    exception.InvalidShareAccessLevel):
                rule_state_map[rule['id']] = {'state': 'error'}
                continue

        for rule in delete_rules:
            self._deny_access('/', share, rule)
        return rule_state_map


class GaneshaNASHelper2(GaneshaNASHelper):
    """Perform share access changes using Ganesha version >= 2.4."""

    def __init__(self, execute, config, tag='<no name>', **kwargs):
        super(GaneshaNASHelper2, self).__init__(execute, config, **kwargs)
        if self.configuration.ganesha_rados_store_enable:
            self.rados_client = kwargs.pop('rados_client')

    def init_helper(self):
        """Initializes protocol-specific NAS drivers."""
        kwargs = {
            'ganesha_config_path': self.configuration.ganesha_config_path,
            'ganesha_export_dir': self.configuration.ganesha_export_dir,
            'ganesha_service_name': self.configuration.ganesha_service_name
        }
        if self.configuration.ganesha_rados_store_enable:
            kwargs['ganesha_rados_store_enable'] = (
                self.configuration.ganesha_rados_store_enable)
            if not self.configuration.ganesha_rados_store_pool_name:
                raise exception.GaneshaException(
                    _('"ganesha_rados_store_pool_name" config option is not '
                      'set in the driver section.'))
            kwargs['ganesha_rados_store_pool_name'] = (
                self.configuration.ganesha_rados_store_pool_name)
            kwargs['ganesha_rados_export_index'] = (
                self.configuration.ganesha_rados_export_index)
            kwargs['ganesha_rados_export_counter'] = (
                self.configuration.ganesha_rados_export_counter)
            kwargs['rados_client'] = self.rados_client
        else:
            kwargs['ganesha_db_path'] = self.configuration.ganesha_db_path
        self.ganesha = ganesha_manager.GaneshaManager(
            self._execute, self.tag, **kwargs)
        system_export_template = self._load_conf_dir(
            self.configuration.ganesha_export_template_dir,
            must_exist=False)
        if system_export_template:
            self.export_template = system_export_template
        else:
            self.export_template = self._default_config_hook()

    def _get_export_path(self, share):
        """Subclass this to return export path."""
        raise NotImplementedError()

    def _get_export_pseudo_path(self, share):
        """Subclass this to return export pseudo path."""
        raise NotImplementedError()

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules of share.

        Creates an export per share. Modifies access rules of shares by
        dynamically updating exports via DBUS.
        """

        confdict = {}
        existing_access_rules = []
        rule_state_map = {}

        if self.ganesha.check_export_exists(share['name']):
            confdict = self.ganesha._read_export(share['name'])
            existing_access_rules = confdict["EXPORT"]["CLIENT"]
            if not isinstance(existing_access_rules, list):
                existing_access_rules = [existing_access_rules]
        else:
            if not access_rules:
                LOG.warning("Trying to remove export file '%s' but it's "
                            "already gone",
                            self.ganesha._getpath(share['name']))
                return

        wanted_rw_clients, wanted_ro_clients = [], []
        for rule in access_rules:

            try:
                ganesha_utils.validate_access_rule(
                    self.supported_access_types, self.supported_access_levels,
                    rule, True)
            except (exception.InvalidShareAccess,
                    exception.InvalidShareAccessLevel):
                rule_state_map[rule['id']] = {'state': 'error'}
                continue

            rule = ganesha_utils.fixup_access_rule(rule)
            if rule['access_level'] == 'rw':
                wanted_rw_clients.append(rule['access_to'])
            elif rule['access_level'] == 'ro':
                wanted_ro_clients.append(rule['access_to'])

        if access_rules:
            # Add or Update export.
            clients = []
            if wanted_ro_clients:
                clients.append({
                    'Access_Type': 'ro',
                    'Clients': ','.join(wanted_ro_clients)
                })
            if wanted_rw_clients:
                clients.append({
                    'Access_Type': 'rw',
                    'Clients': ','.join(wanted_rw_clients)
                })

            if clients:  # Empty list if no rules passed validation
                if existing_access_rules:
                    # Update existing export.
                    ganesha_utils.patch(confdict, {
                        'EXPORT': {
                            'CLIENT': clients
                        }
                    })
                    self.ganesha.update_export(share['name'], confdict)
                else:
                    # Add new export.
                    ganesha_utils.patch(confdict, self.export_template, {
                        'EXPORT': {
                            'Export_Id': self.ganesha.get_export_id(),
                            'Path': self._get_export_path(share),
                            'Pseudo': self._get_export_pseudo_path(share),
                            'Tag': share['name'],
                            'CLIENT': clients,
                            'FSAL': self._fsal_hook(None, share, None)
                        }
                    })
                    self.ganesha.add_export(share['name'], confdict)
        else:
            # No clients have access to the share. Remove export.
            self.ganesha.remove_export(share['name'])
            self._cleanup_fsal_hook(None, share, None)
        return rule_state_map



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import io
import re
from unittest import mock

import ddt
from oslo_serialization import jsonutils

from manila import exception
from manila.share.drivers.ganesha import manager
from manila import test
from manila import utils


test_export_id = 101
test_name = 'fakefile'
test_path = '/fakedir0/export.d/fakefile.conf'
test_tmp_path = '/fakedir0/export.d/fakefile.conf.RANDOM'
test_ganesha_cnf = """EXPORT {
    Export_Id = 101;
    CLIENT {
        Clients = ip1;
        Access_Level = ro;
    }
    CLIENT {
        Clients = ip2;
        Access_Level = rw;
    }
}"""
test_dict_unicode = {
    u'EXPORT': {
        u'Export_Id': 101,
        u'CLIENT': [
            {u'Clients': u"ip1", u'Access_Level': u'ro'},
            {u'Clients': u"ip2", u'Access_Level': u'rw'}]
    }
}
test_dict_str = {
    'EXPORT': {
        'Export_Id': 101,
        'CLIENT': [
            {'Clients': 'ip1', 'Access_Level': 'ro'},
            {'Clients': 'ip2', 'Access_Level': 'rw'}]
    }
}

manager_fake_kwargs = {
    'ganesha_config_path': '/fakedir0/fakeconfig',
    'ganesha_db_path': '/fakedir1/fake.db',
    'ganesha_export_dir': '/fakedir0/export.d',
    'ganesha_service_name': 'ganesha.fakeservice'
}


class MockRadosModule(object):
    """Mocked up version of Ceph's RADOS module."""

    class ObjectNotFound(Exception):
        pass

    class OSError(Exception):
        pass

    class WriteOpCtx():

        def __enter__(self):
            return self

        def __exit__(self, type, msg, traceback):
            pass

        def write_full(self, bytes_to_write):
            pass


@ddt.ddt
class MiscTests(test.TestCase):

    @ddt.data({'import_exc': None},
              {'import_exc': ImportError})
    @ddt.unpack
    def test_setup_rados(self, import_exc):
        manager.rados = None
        with mock.patch.object(
                manager.importutils,
                'import_module',
                side_effect=import_exc) as mock_import_module:
            if import_exc:
                self.assertRaises(
                    exception.ShareBackendException, manager.setup_rados)
            else:
                manager.setup_rados()
                self.assertEqual(mock_import_module.return_value,
                                 manager.rados)
            mock_import_module.assert_called_once_with('rados')


class GaneshaConfigTests(test.TestCase):
    """Tests Ganesha config file format convertor functions."""

    ref_ganesha_cnf = """EXPORT {
    CLIENT {
        Clients = ip1;
        Access_Level = "ro";
    }
    CLIENT {
        Clients = ip2;
        Access_Level = "rw";
    }
    Export_Id = 101;
}"""

    @staticmethod
    def conf_mangle(*confs):
        """A "mangler" for the conf format.

        Its purpose is to transform conf data in a way so that semantically
        equivalent confs yield identical results. Besides this objective
        criteria, we seek a good trade-off between the following
        requirements:
        - low lossiness;
        - low code complexity.
        """
        def _conf_mangle(conf):
            # split to expressions by the delimiter ";"
            # (braces are forced to be treated as expressions
            # by sandwiching them in ";"-s)
            conf = re.sub(r'[{}]', r';\g<0>;', conf).split(';')
            # whitespace-split expressions to tokens with
            # (equality is forced to be treated as token by
            # sandwiching in space)
            conf = map(lambda l: l.replace("=", " = ").split(), conf)
            # get rid of by-product empty lists (derived from superflouous
            # ";"-s that might have crept in due to "sandwiching")
            conf = map(lambda x: x, conf)
            # handle the non-deterministic order of confs
            conf = list(conf)
            conf.sort()
            return conf

        return (_conf_mangle(conf) for conf in confs)

    def test_conf2json(self):
        test_ganesha_cnf_with_comment = """EXPORT {
# fake_export_block
    Export_Id = 101;
    CLIENT {
        Clients = ip1;
    }
}"""
        result_dict_unicode = {
            u'EXPORT': {
                u'CLIENT': {u'Clients': u'ip1'},
                u'Export_Id': 101
            }
        }
        ret = manager._conf2json(test_ganesha_cnf_with_comment)
        self.assertEqual(result_dict_unicode, jsonutils.loads(ret))

    def test_parseconf_ganesha_cnf_input(self):
        ret = manager.parseconf(test_ganesha_cnf)
        self.assertEqual(test_dict_unicode, ret)

    def test_parseconf_json_input(self):
        ret = manager.parseconf(jsonutils.dumps(test_dict_str))
        self.assertEqual(test_dict_unicode, ret)

    def test_dump_to_conf(self):
        ganesha_cnf = io.StringIO()
        manager._dump_to_conf(test_dict_str, ganesha_cnf)
        self.assertEqual(*self.conf_mangle(self.ref_ganesha_cnf,
                                           ganesha_cnf.getvalue()))

    def test_mkconf(self):
        ganesha_cnf = manager.mkconf(test_dict_str)
        self.assertEqual(*self.conf_mangle(self.ref_ganesha_cnf,
                                           ganesha_cnf))


@ddt.ddt
class GaneshaManagerTestCase(test.TestCase):
    """Tests GaneshaManager."""

    def instantiate_ganesha_manager(self, *args, **kwargs):
        ganesha_rados_store_enable = kwargs.get('ganesha_rados_store_enable',
                                                False)
        if ganesha_rados_store_enable:
            with mock.patch.object(
                    manager.GaneshaManager,
                    '_get_rados_object') as self.mock_get_rados_object:
                return manager.GaneshaManager(*args, **kwargs)
        else:
            with mock.patch.object(
                    manager.GaneshaManager,
                    'get_export_id',
                    return_value=100) as self.mock_get_export_id:
                return manager.GaneshaManager(*args, **kwargs)

    def setUp(self):
        super(GaneshaManagerTestCase, self).setUp()
        self._execute = mock.Mock(return_value=('', ''))
        self._rados_client = mock.Mock()
        self._manager = self.instantiate_ganesha_manager(
            self._execute, 'faketag',
            rados_client=self._rados_client,
            **manager_fake_kwargs)
        self._setup_rados = mock.Mock()
        self._execute2 = mock.Mock(return_value=('', ''))
        self.mock_object(manager, 'rados', MockRadosModule)
        self.mock_object(manager, 'setup_rados', self._setup_rados)
        fake_kwargs = copy.copy(manager_fake_kwargs)
        fake_kwargs.update(
            ganesha_rados_store_enable=True,
            ganesha_rados_store_pool_name='fakepool',
            ganesha_rados_export_counter='fakecounter',
            ganesha_rados_export_index='fakeindex',
            rados_client=self._rados_client
        )
        self._manager_with_rados_store = self.instantiate_ganesha_manager(
            self._execute2, 'faketag', **fake_kwargs)
        self.mock_object(utils, 'synchronized',
                         mock.Mock(return_value=lambda f: f))

    def test_init(self):
        self.mock_object(self._manager, 'reset_exports')
        self.mock_object(self._manager, 'restart_service')
        self.assertEqual('/fakedir0/fakeconfig',
                         self._manager.ganesha_config_path)
        self.assertEqual('faketag', self._manager.tag)
        self.assertEqual('/fakedir0/export.d',
                         self._manager.ganesha_export_dir)
        self.assertEqual('/fakedir1/fake.db', self._manager.ganesha_db_path)
        self.assertEqual('ganesha.fakeservice', self._manager.ganesha_service)
        self.assertEqual(
            [mock.call('mkdir', '-p', self._manager.ganesha_export_dir),
             mock.call('mkdir', '-p', '/fakedir1'),
             mock.call('sqlite3', self._manager.ganesha_db_path,
                       'create table ganesha(key varchar(20) primary key, '
                       'value int); insert into ganesha values("exportid", '
                       '100);', run_as_root=False, check_exit_code=False)],
            self._execute.call_args_list)
        self.mock_get_export_id.assert_called_once_with(bump=False)

    def test_init_execute_error_log_message(self):
        fake_args = ('foo', 'bar')

        def raise_exception(*args, **kwargs):
            if args == fake_args:
                raise exception.GaneshaCommandFailure()

        test_execute = mock.Mock(side_effect=raise_exception)
        self.mock_object(manager.LOG, 'error')
        test_manager = self.instantiate_ganesha_manager(
            test_execute, 'faketag', **manager_fake_kwargs)
        self.assertRaises(
            exception.GaneshaCommandFailure,
            test_manager.execute,
            *fake_args, message='fakemsg')
        manager.LOG.error.assert_called_once_with(
            mock.ANY, {'tag': 'faketag', 'msg': 'fakemsg'})

    def test_init_execute_error_no_log_message(self):
        fake_args = ('foo', 'bar')

        def raise_exception(*args, **kwargs):
            if args == fake_args:
                raise exception.GaneshaCommandFailure()

        test_execute = mock.Mock(side_effect=raise_exception)
        self.mock_object(manager.LOG, 'error')
        test_manager = self.instantiate_ganesha_manager(
            test_execute, 'faketag', **manager_fake_kwargs)
        self.assertRaises(
            exception.GaneshaCommandFailure,
            test_manager.execute,
            *fake_args, message='fakemsg', makelog=False)
        self.assertFalse(manager.LOG.error.called)

    @ddt.data(False, True)
    def test_init_with_rados_store_and_export_counter_exists(
            self, counter_exists):
        fake_execute = mock.Mock(return_value=('', ''))
        fake_kwargs = copy.copy(manager_fake_kwargs)
        fake_kwargs.update(
            ganesha_rados_store_enable=True,
            ganesha_rados_store_pool_name='fakepool',
            ganesha_rados_export_counter='fakecounter',
            ganesha_rados_export_index='fakeindex',
            rados_client=self._rados_client
        )
        if counter_exists:
            self.mock_object(
                manager.GaneshaManager, '_get_rados_object', mock.Mock())
        else:
            self.mock_object(
                manager.GaneshaManager, '_get_rados_object',
                mock.Mock(side_effect=MockRadosModule.ObjectNotFound))
        self.mock_object(manager.GaneshaManager, '_put_rados_object')

        test_mgr = manager.GaneshaManager(
            fake_execute, 'faketag', **fake_kwargs)

        self.assertEqual('/fakedir0/fakeconfig', test_mgr.ganesha_config_path)
        self.assertEqual('faketag', test_mgr.tag)
        self.assertEqual('/fakedir0/export.d', test_mgr.ganesha_export_dir)
        self.assertEqual('ganesha.fakeservice', test_mgr.ganesha_service)
        fake_execute.assert_called_once_with(
            'mkdir', '-p', '/fakedir0/export.d')
        self.assertTrue(test_mgr.ganesha_rados_store_enable)
        self.assertEqual('fakepool', test_mgr.ganesha_rados_store_pool_name)
        self.assertEqual('fakecounter', test_mgr.ganesha_rados_export_counter)
        self.assertEqual('fakeindex', test_mgr.ganesha_rados_export_index)
        self.assertEqual(self._rados_client, test_mgr.rados_client)
        self._setup_rados.assert_called_with()
        test_mgr._get_rados_object.assert_called_once_with('fakecounter')
        if counter_exists:
            self.assertFalse(test_mgr._put_rados_object.called)
        else:
            test_mgr._put_rados_object.assert_called_once_with(
                'fakecounter', str(1000))

    def test_ganesha_export_dir(self):
        self.assertEqual(
            '/fakedir0/export.d', self._manager.ganesha_export_dir)

    def test_getpath(self):
        self.assertEqual(
            '/fakedir0/export.d/fakefile.conf',
            self._manager._getpath('fakefile'))

    def test_get_export_rados_object_name(self):
        self.assertEqual(
            'ganesha-export-fakeobj',
            self._manager._get_export_rados_object_name('fakeobj'))

    def test_write_tmp_conf_file(self):
        self.mock_object(manager.pipes, 'quote',
                         mock.Mock(side_effect=['fakedata',
                                                test_tmp_path]))
        test_args = [
            ('mktemp', '-p', '/fakedir0/export.d', '-t',
             'fakefile.conf.XXXXXX'),
            ('sh', '-c', 'echo fakedata > %s' % test_tmp_path)]
        test_kwargs = {
            'message': 'writing %s' % test_tmp_path
        }

        def return_tmpfile(*args, **kwargs):
            if args == test_args[0]:
                return (test_tmp_path + '\n', '')
        self.mock_object(self._manager, 'execute',
                         mock.Mock(side_effect=return_tmpfile))

        ret = self._manager._write_tmp_conf_file(test_path, 'fakedata')

        self._manager.execute.assert_has_calls([
            mock.call(*test_args[0]),
            mock.call(*test_args[1], **test_kwargs)])
        manager.pipes.quote.assert_has_calls([
            mock.call('fakedata'),
            mock.call(test_tmp_path)])
        self.assertEqual(test_tmp_path, ret)

    @ddt.data(True, False)
    def test_write_conf_file_with_mv_error(self, mv_error):
        test_data = 'fakedata'
        test_args = [
            ('mv', test_tmp_path, test_path),
            ('rm', test_tmp_path)]
        self.mock_object(self._manager, '_getpath',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_write_tmp_conf_file',
                         mock.Mock(return_value=test_tmp_path))

        def mock_return(*args, **kwargs):
            if args == test_args[0]:
                if mv_error:
                    raise exception.ProcessExecutionError()
                else:
                    return ('', '')

        self.mock_object(self._manager, 'execute',
                         mock.Mock(side_effect=mock_return))

        if mv_error:
            self.assertRaises(
                exception.ProcessExecutionError,
                self._manager._write_conf_file, test_name, test_data)
        else:
            ret = self._manager._write_conf_file(test_name, test_data)

        self._manager._getpath.assert_called_once_with(test_name)
        self._manager._write_tmp_conf_file.assert_called_once_with(
            test_path, test_data)
        if mv_error:
            self._manager.execute.assert_has_calls([
                mock.call(*test_args[0]),
                mock.call(*test_args[1])])
        else:
            self._manager.execute.assert_has_calls([
                mock.call(*test_args[0])])
            self.assertEqual(test_path, ret)

    def test_mkindex(self):
        test_ls_output = 'INDEX.conf\nfakefile.conf\nfakefile.txt'
        test_index = '%include /fakedir0/export.d/fakefile.conf\n'
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=(test_ls_output, '')))
        self.mock_object(self._manager, '_write_conf_file')
        ret = self._manager._mkindex()
        self._manager.execute.assert_called_once_with(
            'ls', '/fakedir0/export.d', run_as_root=False)
        self._manager._write_conf_file.assert_called_once_with(
            'INDEX', test_index)
        self.assertIsNone(ret)

    def test_read_export_rados_object(self):
        self.mock_object(self._manager_with_rados_store,
                         '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        self.mock_object(self._manager_with_rados_store, '_get_rados_object',
                         mock.Mock(return_value=test_ganesha_cnf))
        self.mock_object(manager, 'parseconf',
                         mock.Mock(return_value=test_dict_unicode))

        ret = self._manager_with_rados_store._read_export_rados_object(
            test_name)

        (self._manager_with_rados_store._get_export_rados_object_name.
         assert_called_once_with(test_name))
        (self._manager_with_rados_store._get_rados_object.
         assert_called_once_with('fakeobj'))
        manager.parseconf.assert_called_once_with(test_ganesha_cnf)
        self.assertEqual(test_dict_unicode, ret)

    def test_read_export_file(self):
        test_args = ('cat', test_path)
        test_kwargs = {'message': 'reading export fakefile'}
        self.mock_object(self._manager, '_getpath',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=(test_ganesha_cnf,)))
        self.mock_object(manager, 'parseconf',
                         mock.Mock(return_value=test_dict_unicode))
        ret = self._manager._read_export_file(test_name)
        self._manager._getpath.assert_called_once_with(test_name)
        self._manager.execute.assert_called_once_with(
            *test_args, **test_kwargs)
        manager.parseconf.assert_called_once_with(test_ganesha_cnf)
        self.assertEqual(test_dict_unicode, ret)

    @ddt.data(False, True)
    def test_read_export_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_read_export_file',
                         mock.Mock(return_value=test_dict_unicode))
        self.mock_object(self._manager, '_read_export_rados_object',
                         mock.Mock(return_value=test_dict_unicode))

        ret = self._manager._read_export(test_name)

        if rados_store_enable:
            self._manager._read_export_rados_object.assert_called_once_with(
                test_name)
            self.assertFalse(self._manager._read_export_file.called)
        else:
            self._manager._read_export_file.assert_called_once_with(test_name)
            self.assertFalse(self._manager._read_export_rados_object.called)
        self.assertEqual(test_dict_unicode, ret)

    @ddt.data(True, False)
    def test_check_export_rados_object_exists(self, exists):
        self.mock_object(
            self._manager_with_rados_store,
            '_get_export_rados_object_name', mock.Mock(return_value='fakeobj'))
        if exists:
            self.mock_object(
                self._manager_with_rados_store, '_get_rados_object')
        else:
            self.mock_object(
                self._manager_with_rados_store, '_get_rados_object',
                mock.Mock(side_effect=MockRadosModule.ObjectNotFound))

        ret = self._manager_with_rados_store._check_export_rados_object_exists(
            test_name)

        (self._manager_with_rados_store._get_export_rados_object_name.
         assert_called_once_with(test_name))
        (self._manager_with_rados_store._get_rados_object.
         assert_called_once_with('fakeobj'))
        if exists:
            self.assertTrue(ret)
        else:
            self.assertFalse(ret)

    def test_check_file_exists(self):
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=(test_ganesha_cnf,)))

        ret = self._manager._check_file_exists(test_path)

        self._manager.execute.assert_called_once_with(
            'test', '-f', test_path, makelog=False, run_as_root=False)
        self.assertTrue(ret)

    @ddt.data(1, 4)
    def test_check_file_exists_error(self, exit_code):
        self.mock_object(
            self._manager, 'execute',
            mock.Mock(side_effect=exception.GaneshaCommandFailure(
                exit_code=exit_code))
        )

        if exit_code == 1:
            ret = self._manager._check_file_exists(test_path)
            self.assertFalse(ret)
        else:
            self.assertRaises(exception.GaneshaCommandFailure,
                              self._manager._check_file_exists,
                              test_path)

        self._manager.execute.assert_called_once_with(
            'test', '-f', test_path, makelog=False, run_as_root=False)

    def test_check_export_file_exists(self):
        self.mock_object(self._manager, '_getpath',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_check_file_exists',
                         mock.Mock(return_value=True))

        ret = self._manager._check_export_file_exists(test_name)

        self._manager._getpath.assert_called_once_with(test_name)
        self._manager._check_file_exists.assert_called_once_with(test_path)
        self.assertTrue(ret)

    @ddt.data(False, True)
    def test_check_export_exists_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_check_export_file_exists',
                         mock.Mock(return_value=True))
        self.mock_object(self._manager, '_check_export_rados_object_exists',
                         mock.Mock(return_value=True))

        ret = self._manager.check_export_exists(test_name)

        if rados_store_enable:
            (self._manager._check_export_rados_object_exists.
             assert_called_once_with(test_name))
            self.assertFalse(self._manager._check_export_file_exists.called)
        else:
            self._manager._check_export_file_exists.assert_called_once_with(
                test_name)
            self.assertFalse(
                self._manager._check_export_rados_object_exists.called)
        self.assertTrue(ret)

    def test_write_export_rados_object(self):
        self.mock_object(self._manager, '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        self.mock_object(self._manager, '_put_rados_object')
        self.mock_object(self._manager, '_getpath',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_write_tmp_conf_file',
                         mock.Mock(return_value=test_tmp_path))

        ret = self._manager._write_export_rados_object(test_name, 'fakedata')

        self._manager._get_export_rados_object_name.assert_called_once_with(
            test_name)
        self._manager._put_rados_object.assert_called_once_with(
            'fakeobj', 'fakedata')
        self._manager._getpath.assert_called_once_with(test_name)
        self._manager._write_tmp_conf_file.assert_called_once_with(
            test_path, 'fakedata')
        self.assertEqual(test_tmp_path, ret)

    @ddt.data(True, False)
    def test_write_export_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(manager, 'mkconf',
                         mock.Mock(return_value=test_ganesha_cnf))
        self.mock_object(self._manager, '_write_conf_file',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_write_export_rados_object',
                         mock.Mock(return_value=test_path))

        ret = self._manager._write_export(test_name, test_dict_str)

        manager.mkconf.assert_called_once_with(test_dict_str)
        if rados_store_enable:
            self._manager._write_export_rados_object.assert_called_once_with(
                test_name, test_ganesha_cnf)
            self.assertFalse(self._manager._write_conf_file.called)
        else:
            self._manager._write_conf_file.assert_called_once_with(
                test_name, test_ganesha_cnf)
            self.assertFalse(self._manager._write_export_rados_object.called)
        self.assertEqual(test_path, ret)

    def test_write_export_error_incomplete_export_block(self):
        test_errordict = {
            u'EXPORT': {
                u'Export_Id': '@config',
                u'CLIENT': {u'Clients': u"'ip1','ip2'"}
            }
        }
        self.mock_object(manager, 'mkconf',
                         mock.Mock(return_value=test_ganesha_cnf))
        self.mock_object(self._manager, '_write_conf_file',
                         mock.Mock(return_value=test_path))

        self.assertRaises(exception.InvalidParameterValue,
                          self._manager._write_export,
                          test_name, test_errordict)

        self.assertFalse(manager.mkconf.called)
        self.assertFalse(self._manager._write_conf_file.called)

    def test_rm_file(self):
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=('', '')))
        ret = self._manager._rm_export_file(test_name)

        self._manager.execute.assert_called_once_with('rm', '-f', test_path)
        self.assertIsNone(ret)

    def test_rm_export_file(self):
        self.mock_object(self._manager, '_getpath',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_rm_file')

        ret = self._manager._rm_export_file(test_name)

        self._manager._getpath.assert_called_once_with(test_name)
        self._manager._rm_file.assert_called_once_with(test_path)
        self.assertIsNone(ret)

    def test_rm_export_rados_object(self):
        self.mock_object(self._manager_with_rados_store,
                         '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        self.mock_object(self._manager_with_rados_store,
                         '_delete_rados_object')

        ret = self._manager_with_rados_store._rm_export_rados_object(
            test_name)

        (self._manager_with_rados_store._get_export_rados_object_name.
         assert_called_once_with(test_name))
        (self._manager_with_rados_store._delete_rados_object.
         assert_called_once_with('fakeobj'))
        self.assertIsNone(ret)

    def test_dbus_send_ganesha(self):
        test_args = ('arg1', 'arg2')
        test_kwargs = {'key': 'value'}
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=('', '')))
        ret = self._manager._dbus_send_ganesha('fakemethod', *test_args,
                                               **test_kwargs)
        self._manager.execute.assert_called_once_with(
            'dbus-send', '--print-reply', '--system',
            '--dest=org.ganesha.nfsd', '/org/ganesha/nfsd/ExportMgr',
            'org.ganesha.nfsd.exportmgr.fakemethod',
            *test_args, message='dbus call exportmgr.fakemethod',
            **test_kwargs)
        self.assertIsNone(ret)

    def test_remove_export_dbus(self):
        self.mock_object(self._manager, '_dbus_send_ganesha')
        ret = self._manager._remove_export_dbus(test_export_id)
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'RemoveExport', 'uint16:101')
        self.assertIsNone(ret)

    @ddt.data('',
              '%url rados://fakepool/fakeobj2')
    def test_add_rados_object_url_to_index_with_index_data(
            self, index_data):
        self.mock_object(
            self._manager_with_rados_store, '_get_rados_object',
            mock.Mock(return_value=index_data))
        self.mock_object(
            self._manager_with_rados_store, '_get_export_rados_object_name',
            mock.Mock(return_value='fakeobj1'))
        self.mock_object(
            self._manager_with_rados_store, '_put_rados_object')

        ret = (self._manager_with_rados_store.
               _add_rados_object_url_to_index('fakename'))

        (self._manager_with_rados_store._get_rados_object.
         assert_called_once_with('fakeindex'))
        (self._manager_with_rados_store._get_export_rados_object_name.
         assert_called_once_with('fakename'))
        if index_data:
            urls = ('%url rados://fakepool/fakeobj2\n'
                    '%url rados://fakepool/fakeobj1')
        else:
            urls = '%url rados://fakepool/fakeobj1'
        (self._manager_with_rados_store._put_rados_object.
         assert_called_once_with('fakeindex', urls))
        self.assertIsNone(ret)

    @ddt.data('',
              '%url rados://fakepool/fakeobj1\n'
              '%url rados://fakepool/fakeobj2')
    def test_remove_rados_object_url_from_index_with_index_data(
            self, index_data):
        self.mock_object(
            self._manager_with_rados_store, '_get_rados_object',
            mock.Mock(return_value=index_data))
        self.mock_object(
            self._manager_with_rados_store, '_get_export_rados_object_name',
            mock.Mock(return_value='fakeobj1'))
        self.mock_object(
            self._manager_with_rados_store, '_put_rados_object')

        ret = (self._manager_with_rados_store.
               _remove_rados_object_url_from_index('fakename'))

        if index_data:
            (self._manager_with_rados_store._get_rados_object.
             assert_called_once_with('fakeindex'))
            (self._manager_with_rados_store._get_export_rados_object_name.
             assert_called_once_with('fakename'))
            urls = '%url rados://fakepool/fakeobj2'
            (self._manager_with_rados_store._put_rados_object.
             assert_called_once_with('fakeindex', urls))
        else:
            (self._manager_with_rados_store._get_rados_object.
             assert_called_once_with('fakeindex'))
            self.assertFalse(self._manager_with_rados_store.
                             _get_export_rados_object_name.called)
            self.assertFalse(self._manager_with_rados_store.
                             _put_rados_object.called)
        self.assertIsNone(ret)

    @ddt.data(False, True)
    def test_add_export_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_write_export',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_dbus_send_ganesha')
        self.mock_object(self._manager, '_rm_file')
        self.mock_object(self._manager, '_add_rados_object_url_to_index')
        self.mock_object(self._manager, '_mkindex')

        ret = self._manager.add_export(test_name, test_dict_str)

        self._manager._write_export.assert_called_once_with(
            test_name, test_dict_str)
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'AddExport', 'string:' + test_path,
            'string:EXPORT(Export_Id=101)')
        if rados_store_enable:
            self._manager._rm_file.assert_called_once_with(test_path)
            self._manager._add_rados_object_url_to_index(test_name)
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._mkindex.assert_called_once_with()
            self.assertFalse(self._manager._rm_file.called)
            self.assertFalse(
                self._manager._add_rados_object_url_to_index.called)
        self.assertIsNone(ret)

    def test_add_export_error_during_mkindex(self):
        self.mock_object(self._manager, '_write_export',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_dbus_send_ganesha')
        self.mock_object(
            self._manager, '_mkindex',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        self.mock_object(self._manager, '_rm_export_file')
        self.mock_object(self._manager, '_remove_export_dbus')

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.add_export, test_name, test_dict_str)

        self._manager._write_export.assert_called_once_with(
            test_name, test_dict_str)
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'AddExport', 'string:' + test_path,
            'string:EXPORT(Export_Id=101)')
        self._manager._mkindex.assert_called_once_with()
        self._manager._rm_export_file.assert_called_once_with(test_name)
        self._manager._remove_export_dbus.assert_called_once_with(
            test_export_id)

    @ddt.data(True, False)
    def test_add_export_error_during_write_export_with_rados_store(
            self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(
            self._manager, '_write_export',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        self.mock_object(self._manager, '_mkindex')

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.add_export, test_name, test_dict_str)

        self._manager._write_export.assert_called_once_with(
            test_name, test_dict_str)
        if rados_store_enable:
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._mkindex.assert_called_once_with()

    @ddt.data(True, False)
    def test_add_export_error_during_dbus_send_ganesha_with_rados_store(
            self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_write_export',
                         mock.Mock(return_value=test_path))
        self.mock_object(
            self._manager, '_dbus_send_ganesha',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        self.mock_object(self._manager, '_mkindex')
        self.mock_object(self._manager, '_rm_export_file')
        self.mock_object(self._manager, '_rm_export_rados_object')
        self.mock_object(self._manager, '_rm_file')
        self.mock_object(self._manager, '_remove_export_dbus')

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.add_export, test_name, test_dict_str)

        self._manager._write_export.assert_called_once_with(
            test_name, test_dict_str)
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'AddExport', 'string:' + test_path,
            'string:EXPORT(Export_Id=101)')
        if rados_store_enable:
            self._manager._rm_export_rados_object.assert_called_once_with(
                test_name)
            self._manager._rm_file.assert_called_once_with(test_path)
            self.assertFalse(self._manager._rm_export_file.called)
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._rm_export_file.assert_called_once_with(test_name)
            self._manager._mkindex.assert_called_once_with()
            self.assertFalse(self._manager._rm_export_rados_object.called)
            self.assertFalse(self._manager._rm_file.called)
        self.assertFalse(self._manager._remove_export_dbus.called)

    @ddt.data(True, False)
    def test_update_export_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        confdict = {
            'EXPORT': {
                'Export_Id': 101,
                'CLIENT': {'Clients': 'ip1', 'Access_Level': 'ro'},
            }
        }
        self.mock_object(self._manager, '_read_export',
                         mock.Mock(return_value=test_dict_unicode))
        self.mock_object(self._manager, '_write_export',
                         mock.Mock(return_value=test_path))
        self.mock_object(self._manager, '_dbus_send_ganesha')
        self.mock_object(self._manager, '_rm_file')

        self._manager.update_export(test_name, confdict)

        self._manager._read_export.assert_called_once_with(test_name)
        self._manager._write_export.assert_called_once_with(test_name,
                                                            confdict)
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'UpdateExport', 'string:' + test_path,
            'string:EXPORT(Export_Id=101)')
        if rados_store_enable:
            self._manager._rm_file.assert_called_once_with(test_path)
        else:
            self.assertFalse(self._manager._rm_file.called)

    @ddt.data(True, False)
    def test_update_export_error_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        confdict = {
            'EXPORT': {
                'Export_Id': 101,
                'CLIENT': {'Clients': 'ip1', 'Access_Level': 'ro'},
            }
        }
        self.mock_object(self._manager, '_read_export',
                         mock.Mock(return_value=test_dict_unicode))
        self.mock_object(self._manager, '_write_export',
                         mock.Mock(return_value=test_path))
        self.mock_object(
            self._manager, '_dbus_send_ganesha',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        self.mock_object(self._manager, '_rm_file')

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.update_export, test_name, confdict)

        self._manager._read_export.assert_called_once_with(test_name)
        self._manager._write_export.assert_has_calls([
            mock.call(test_name, confdict),
            mock.call(test_name, test_dict_unicode)])
        self._manager._dbus_send_ganesha.assert_called_once_with(
            'UpdateExport', 'string:' + test_path,
            'string:EXPORT(Export_Id=101)')
        if rados_store_enable:
            self._manager._rm_file.assert_called_once_with(test_path)
        else:
            self.assertFalse(self._manager._rm_file.called)

    @ddt.data(True, False)
    def test_remove_export_with_rados_store(self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_read_export',
                         mock.Mock(return_value=test_dict_unicode))
        self.mock_object(self._manager, '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        methods = ('_remove_export_dbus', '_rm_export_file', '_mkindex',
                   '_remove_rados_object_url_from_index',
                   '_delete_rados_object')
        for method in methods:
            self.mock_object(self._manager, method)

        ret = self._manager.remove_export(test_name)

        self._manager._read_export.assert_called_once_with(test_name)
        self._manager._remove_export_dbus.assert_called_once_with(
            test_dict_unicode['EXPORT']['Export_Id'])
        if rados_store_enable:
            (self._manager._get_export_rados_object_name.
             assert_called_once_with(test_name))
            self._manager._delete_rados_object.assert_called_once_with(
                'fakeobj')
            (self._manager._remove_rados_object_url_from_index.
             assert_called_once_with(test_name))
            self.assertFalse(self._manager._rm_export_file.called)
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._rm_export_file.assert_called_once_with(test_name)
            self._manager._mkindex.assert_called_once_with()
            self.assertFalse(
                self._manager._get_export_rados_object_name.called)
            self.assertFalse(self._manager._delete_rados_object.called)
            self.assertFalse(
                self._manager._remove_rados_object_url_from_index.called)
        self.assertIsNone(ret)

    @ddt.data(True, False)
    def test_remove_export_error_during_read_export_with_rados_store(
            self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(
            self._manager, '_read_export',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        self.mock_object(self._manager, '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        methods = ('_remove_export_dbus', '_rm_export_file', '_mkindex',
                   '_remove_rados_object_url_from_index',
                   '_delete_rados_object')
        for method in methods:
            self.mock_object(self._manager, method)

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.remove_export, test_name)

        self._manager._read_export.assert_called_once_with(test_name)
        self.assertFalse(self._manager._remove_export_dbus.called)
        if rados_store_enable:
            (self._manager._get_export_rados_object_name.
             assert_called_once_with(test_name))
            self._manager._delete_rados_object.assert_called_once_with(
                'fakeobj')
            (self._manager._remove_rados_object_url_from_index.
             assert_called_once_with(test_name))
            self.assertFalse(self._manager._rm_export_file.called)
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._rm_export_file.assert_called_once_with(test_name)
            self._manager._mkindex.assert_called_once_with()
            self.assertFalse(
                self._manager._get_export_rados_object_name.called)
            self.assertFalse(self._manager._delete_rados_object.called)
            self.assertFalse(
                self._manager._remove_rados_object_url_from_index.called)

    @ddt.data(True, False)
    def test_remove_export_error_during_remove_export_dbus_with_rados_store(
            self, rados_store_enable):
        self._manager.ganesha_rados_store_enable = rados_store_enable
        self.mock_object(self._manager, '_read_export',
                         mock.Mock(return_value=test_dict_unicode))
        self.mock_object(self._manager, '_get_export_rados_object_name',
                         mock.Mock(return_value='fakeobj'))
        self.mock_object(
            self._manager, '_remove_export_dbus',
            mock.Mock(side_effect=exception.GaneshaCommandFailure))
        methods = ('_rm_export_file', '_mkindex',
                   '_remove_rados_object_url_from_index',
                   '_delete_rados_object')
        for method in methods:
            self.mock_object(self._manager, method)

        self.assertRaises(exception.GaneshaCommandFailure,
                          self._manager.remove_export, test_name)

        self._manager._read_export.assert_called_once_with(test_name)
        self._manager._remove_export_dbus.assert_called_once_with(
            test_dict_unicode['EXPORT']['Export_Id'])
        if rados_store_enable:
            (self._manager._get_export_rados_object_name.
             assert_called_once_with(test_name))
            self._manager._delete_rados_object.assert_called_once_with(
                'fakeobj')
            (self._manager._remove_rados_object_url_from_index.
             assert_called_once_with(test_name))
            self.assertFalse(self._manager._rm_export_file.called)
            self.assertFalse(self._manager._mkindex.called)
        else:
            self._manager._rm_export_file.assert_called_once_with(test_name)
            self._manager._mkindex.assert_called_once_with()
            self.assertFalse(
                self._manager._get_export_rados_object_name.called)
            self.assertFalse(self._manager._delete_rados_object.called)
            self.assertFalse(
                self._manager._remove_rados_object_url_from_index.called)

    def test_get_rados_object(self):
        fakebin = chr(246).encode('utf-8')

        ioctx = mock.Mock()
        ioctx.read.side_effect = [fakebin, fakebin]

        self._rados_client.open_ioctx = mock.Mock(return_value=ioctx)
        self._rados_client.conf_get = mock.Mock(return_value=256)

        max_size = 256 * 1024 * 1024

        ret = self._manager_with_rados_store._get_rados_object('fakeobj')

        self._rados_client.open_ioctx.assert_called_once_with('fakepool')
        self._rados_client.conf_get.assert_called_once_with(
            'osd_max_write_size')
        ioctx.read.assert_called_once_with('fakeobj', max_size)
        ioctx.close.assert_called_once()

        self.assertEqual(fakebin.decode('utf-8'), ret)

    def test_put_rados_object(self):
        faketext = chr(246)

        ioctx = mock.Mock()
        manager.rados.WriteOpCtx.write_full = mock.Mock()

        self._rados_client.open_ioctx = mock.Mock(return_value=ioctx)
        self._rados_client.conf_get = mock.Mock(return_value=256)

        ret = self._manager_with_rados_store._put_rados_object(
            'fakeobj', faketext)

        self._rados_client.open_ioctx.assert_called_once_with('fakepool')
        self._rados_client.conf_get.assert_called_once_with(
            'osd_max_write_size')
        manager.rados.WriteOpCtx.write_full.assert_called_once_with(
            faketext.encode('utf-8'))
        ioctx.operate_write_op.assert_called_once_with(mock.ANY, 'fakeobj')

        self.assertIsNone(ret)

    def test_delete_rados_object(self):
        ioctx = mock.Mock()

        self._rados_client.open_ioctx = mock.Mock(return_value=ioctx)

        ret = self._manager_with_rados_store._delete_rados_object('fakeobj')

        self._rados_client.open_ioctx.assert_called_once_with('fakepool')
        ioctx.remove_object.assert_called_once_with('fakeobj')
        ioctx.close.assert_called_once()

        self.assertIsNone(ret)

    def test_get_export_id(self):
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=('exportid|101', '')))
        ret = self._manager.get_export_id()
        self._manager.execute.assert_called_once_with(
            'sqlite3', self._manager.ganesha_db_path,
            'update ganesha set value = value + 1;'
            'select * from ganesha where key = "exportid";',
            run_as_root=False)
        self.assertEqual(101, ret)

    def test_get_export_id_nobump(self):
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=('exportid|101', '')))
        ret = self._manager.get_export_id(bump=False)
        self._manager.execute.assert_called_once_with(
            'sqlite3', self._manager.ganesha_db_path,
            'select * from ganesha where key = "exportid";',
            run_as_root=False)
        self.assertEqual(101, ret)

    def test_get_export_id_error_invalid_export_db(self):
        self.mock_object(self._manager, 'execute',
                         mock.Mock(return_value=('invalid', '')))
        self.mock_object(manager.LOG, 'error')
        self.assertRaises(exception.InvalidSqliteDB,
                          self._manager.get_export_id)
        manager.LOG.error.assert_called_once_with(
            mock.ANY, mock.ANY)
        self._manager.execute.assert_called_once_with(
            'sqlite3', self._manager.ganesha_db_path,
            'update ganesha set value = value + 1;'
            'select * from ganesha where key = "exportid";',
            run_as_root=False)

    @ddt.data(True, False)
    def test_get_export_id_with_rados_store_and_bump(self, bump):
        self.mock_object(self._manager_with_rados_store,
                         '_get_rados_object', mock.Mock(return_value='1000'))
        self.mock_object(self._manager_with_rados_store, '_put_rados_object')

        ret = self._manager_with_rados_store.get_export_id(bump=bump)

        if bump:
            (self._manager_with_rados_store._get_rados_object.
             assert_called_once_with('fakecounter'))
            (self._manager_with_rados_store._put_rados_object.
             assert_called_once_with('fakecounter', '1001'))
            self.assertEqual(1001, ret)
        else:
            (self._manager_with_rados_store._get_rados_object.
             assert_called_once_with('fakecounter'))
            self.assertFalse(
                self._manager_with_rados_store._put_rados_object.called)
            self.assertEqual(1000, ret)

    def test_restart_service(self):
        self.mock_object(self._manager, 'execute')
        ret = self._manager.restart_service()
        self._manager.execute.assert_called_once_with(
            'service', 'ganesha.fakeservice', 'restart')
        self.assertIsNone(ret)

    def test_reset_exports(self):
        self.mock_object(self._manager, 'execute')
        self.mock_object(self._manager, '_mkindex')
        ret = self._manager.reset_exports()
        self._manager.execute.assert_called_once_with(
            'sh', '-c', 'rm -f /fakedir0/export.d/*.conf')
        self._manager._mkindex.assert_called_once_with()
        self.assertIsNone(ret)



# Copyright (c) 2015 Hitachi Data Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Helper class for Share Migration."""

import time

from oslo_config import cfg
from oslo_log import log

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import api as share_api
from manila.share import rpcapi as share_rpcapi
import manila.utils as utils


LOG = log.getLogger(__name__)

migration_opts = [
    cfg.IntOpt(
        'migration_wait_access_rules_timeout',
        default=180,
        help="Time to wait for access rules to be allowed/denied on backends "
             "when migrating shares using generic approach (seconds)."),
    cfg.IntOpt(
        'migration_create_delete_share_timeout',
        default=300,
        help='Timeout for creating and deleting share instances '
             'when performing share migration (seconds).'),
]

CONF = cfg.CONF
CONF.register_opts(migration_opts)


class ShareMigrationHelper(object):

    def __init__(self, context, db, access_helper):

        self.db = db
        self.context = context
        self.access_helper = access_helper
        self.api = share_api.API()
        self.access_helper = access_helper

        self.migration_create_delete_share_timeout = (
            CONF.migration_create_delete_share_timeout)
        self.migration_wait_access_rules_timeout = (
            CONF.migration_wait_access_rules_timeout)

    def delete_instance_and_wait(self, share_instance):

        self.api.delete_instance(self.context, share_instance, True)

        # Wait for deletion.
        starttime = time.time()
        deadline = starttime + self.migration_create_delete_share_timeout
        tries = 0
        instance = "Something not None"
        while instance is not None:
            try:
                instance = self.db.share_instance_get(self.context,
                                                      share_instance['id'])
                tries += 1
                now = time.time()
                if now > deadline:
                    msg = _("Timeout trying to delete instance "
                            "%s") % share_instance['id']
                    raise exception.ShareMigrationFailed(reason=msg)
            except exception.NotFound:
                instance = None
            else:
                # 1.414 = square-root of 2
                time.sleep(1.414 ** tries)

    def create_instance_and_wait(self, share, dest_host, new_share_network_id,
                                 new_az_id, new_share_type_id):

        new_share_instance = self.api.create_instance(
            self.context, share, new_share_network_id, dest_host,
            new_az_id, share_type_id=new_share_type_id)

        # Wait for new_share_instance to become ready
        starttime = time.time()
        deadline = starttime + self.migration_create_delete_share_timeout
        new_share_instance = self.db.share_instance_get(
            self.context, new_share_instance['id'], with_share_data=True)
        tries = 0
        while new_share_instance['status'] != constants.STATUS_AVAILABLE:
            tries += 1
            now = time.time()
            if new_share_instance['status'] == constants.STATUS_ERROR:
                msg = _("Failed to create new share instance"
                        " (from %(share_id)s) on "
                        "destination host %(host_name)s") % {
                    'share_id': share['id'], 'host_name': dest_host}
                self.cleanup_new_instance(new_share_instance)
                raise exception.ShareMigrationFailed(reason=msg)
            elif now > deadline:
                msg = _("Timeout creating new share instance "
                        "(from %(share_id)s) on "
                        "destination host %(host_name)s") % {
                    'share_id': share['id'], 'host_name': dest_host}
                self.cleanup_new_instance(new_share_instance)
                raise exception.ShareMigrationFailed(reason=msg)
            else:
                # 1.414 = square-root of 2
                time.sleep(1.414 ** tries)
            new_share_instance = self.db.share_instance_get(
                self.context, new_share_instance['id'], with_share_data=True)

        return new_share_instance

    # NOTE(ganso): Cleanup methods do not throw exceptions, since the
    # exceptions that should be thrown are the ones that call the cleanup

    def cleanup_new_instance(self, new_instance):

        try:
            self.delete_instance_and_wait(new_instance)
        except Exception:
            LOG.warning("Failed to cleanup new instance during generic "
                        "migration for share %s.", new_instance['share_id'])

    def cleanup_access_rules(self, share_instances, share_server,
                             dest_host=None):

        try:
            self.revert_access_rules(share_instances, share_server, dest_host)
        except Exception:
            LOG.warning("Failed to cleanup access rules during generic"
                        " migration.")

    def revert_access_rules(self, share_instances, share_server,
                            dest_host=None):
        shares_instance_ids = []
        for share_instance in share_instances:
            # Cast all rules to 'queued_to_apply' so that they can be
            # re-applied.
            shares_instance_ids.append(share_instance['id'])
            updates = {'state': constants.ACCESS_STATE_QUEUED_TO_APPLY}
            self.access_helper.get_and_update_share_instance_access_rules(
                self.context, updates=updates,
                share_instance_id=share_instance['id'])

        if dest_host:
            rpcapi = share_rpcapi.ShareAPI()
            rpcapi.update_access_for_instances(self.context, dest_host,
                                               shares_instance_ids,
                                               share_server)
        else:
            for share_instance in share_instances:
                self.access_helper.update_access_rules(
                    self.context, share_instance['id'],
                    share_server=share_server)

        for share_instance in share_instances:
            utils.wait_for_access_update(
                self.context, self.db, share_instance,
                self.migration_wait_access_rules_timeout)

    def apply_new_access_rules(self, new_share_instance, share_id):

        rules = self.db.share_instance_access_copy(
            self.context, share_id, new_share_instance['id'])

        if rules:
            self.api.allow_access_to_instance(self.context, new_share_instance)

            utils.wait_for_access_update(
                self.context, self.db, new_share_instance,
                self.migration_wait_access_rules_timeout)
        else:
            LOG.debug("No access rules to sync to destination share instance.")

    @utils.retry(retry_param=exception.ShareServerNotReady, retries=8)
    def wait_for_share_server(self, share_server_id):
        share_server = self.db.share_server_get(self.context, share_server_id)
        if share_server['status'] == constants.STATUS_ERROR:
            raise exception.ShareServerNotCreated(
                share_server_id=share_server_id)
        elif share_server['status'] == constants.STATUS_ACTIVE:
            return share_server
        else:
            raise exception.ShareServerNotReady(
                share_server_id=share_server_id, time=511,
                state=constants.STATUS_AVAILABLE)
